"""becquerel package metadata."""

__name__ = "becquerel"
__author__ = "The Becquerel Development Team"
__maintainer__ = __author__
__email__ = "becquerel-dev@lbl.gov"
__description__ = "Tools for radiation spectral analysis."
__url__ = "https://github.com/lbl-anp/becquerel"
__version__ = "0.4.5"
# classifiers from list at https://pypi.org/classifiers/
__classifiers__ = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Science/Research",
    "License :: Other/Proprietary License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Physics",
]
__license__ = "Other/Proprietary License (see LICENSE.txt)"
__copyright__ = """\
becquerel (bq) Copyright (c) 2017-2021, The Regents of the University of
California, through Lawrence Berkeley National Laboratory (subject to receipt
of any required approvals from the U.S. Dept. of Energy) and University of
California, Berkeley.  All rights reserved.

If you have questions about your rights to use or distribute this software,
please contact Berkeley Lab's Intellectual Property Office at
IPO@lbl.gov.

NOTICE.  This Software was developed under funding from the U.S. Department
of Energy and the U.S. Government consequently retains certain rights.  As
such, the U.S. Government has been granted for itself and others acting on
its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
Software to reproduce, distribute copies to the public, prepare derivative
works, and perform publicly and display publicly, and to permit others to do so.
"""



"""Read in an Ortec SPC file."""

import os
import struct
import dateutil.parser
import numpy as np
from ..core import calibration
from .parsers import BecquerelParserError


SPC_FORMAT_BEGINNING = [
    # Record 1
    [
        ["INFTYP", "H"],
        ["FILTYP", "H"],
        ["3", "H"],
        ["4", "H"],
        ["ACQIRP", "H"],
        ["SAMDRP", "H"],
        ["DETDRP", "H"],
        ["EBRDESC", "H"],
        ["ANARP1", "H"],
        ["ANARP2", "H"],
        ["ANARP3", "H"],
        ["ANARP4", "H"],
        ["SRPDES", "H"],
        ["IEQDESC", "H"],
        ["GEODES", "H"],
        ["MPCDESC", "H"],
        ["CALDES", "H"],
        ["CALRP1", "H"],
        ["CALRP2", "H"],
        ["EFFPRP", "H"],
        ["ROIRP1", "H"],
        ["22", "H"],
        ["23", "H"],
        ["24", "H"],
        ["25", "H"],
        ["26", "H"],
        ["PERPTR", "H"],
        ["MAXRCS", "H"],
        ["LSTREC", "H"],
        ["EFFPNM", "H"],
        ["SPCTRP", "H"],
        ["SPCRCN", "H"],
        ["SPCCHN", "H"],
        ["ABSTCH", "H"],
        ["ACQTIM", "f"],
        ["ACQTI8", "d"],
        ["SEQNUM", "H"],
        ["MCANU", "H"],
        ["SEGNUM", "H"],
        ["MCADVT", "H"],
        ["CHNSRT", "H"],
        ["RLTMDT", "f"],
        ["LVTMDT", "f"],
        ["50", "H"],
        ["51", "H"],
        ["52", "H"],
        ["53", "H"],
        ["54", "H"],
        ["55", "H"],
        ["56", "H"],
        ["57", "H"],
        ["58", "H"],
        ["59", "H"],
        ["60", "H"],
        ["61", "H"],
        ["62", "H"],
        ["RRSFCT", "f"],
    ],
    # Unknown Record
    [
        ["Unknown Record 1", "128B"],
    ],
    # Acquisition Information Record
    [
        ["Default spectrum file name", "16s"],
        ["Date", "12s"],
        ["Time", "10s"],
        ["Live Time", "10s"],
        ["Real Time", "10s"],
        ["59--90", "B33x"],
        ["Start date of sample collection", "10s"],
        ["Start time of sample collection", "8s"],
        ["Stop date of sample collection", "10s"],
        ["Stop time of sample collection", "8s"],
    ],
    # Sample Description Record
    [
        ["Sample Description", "128s"],
    ],
    # Detector Description Record
    [
        ["Detector Description", "128s"],
    ],
    # First Analysis Parameter
    [
        ["Calibration ?", "16f"],
        ["Testing", "64s"],
    ],
    # Unknown Record
    [
        ["Unknown Record 2", "128B"],
    ],
    # Calibration Description Record
    [
        ["Calibration Description", "128s"],
    ],
    # Description Record 1
    [
        ["Location Description Record 1", "x127s"],
    ],
    # Description Record 2
    [
        ["Location Description Record 2", "128s"],
    ],
    # Unknown Record
    [
        ["Unknown Record 3", "128B"],
    ],
    # Unknown Record
    [
        ["Unknown Record 4", "128B"],
    ],
    # Unknown Record
    [
        ["Unknown Record 5", "128B"],
    ],
    # Unknown Record
    [
        ["Unknown Record 6", "128B"],
    ],
    # Empty Record
    [
        ["Empty Record 1", "128B"],
    ],
    # Empty Record
    [
        ["Empty Record 2", "128B"],
    ],
    # Empty Record
    [
        ["Empty Record 3", "128B"],
    ],
    # Empty Record
    [
        ["Empty Record 4", "128B"],
    ],
    # Empty Record
    [
        ["Empty Record 5", "128B"],
    ],
    # Hardware Parameters Record 1
    [
        ["Hardware Parameters Record 1", "128s"],
    ],
    # Hardware Parameters Record 2
    [
        ["Hardware Parameters Record 2", "128s"],
    ],
]

SPC_FORMAT_END = [
    # Unknown Record
    [
        ["Unknown Record 1", "128B"],
    ],
    # Unknown Record
    [
        ["Unknown Record 2", "128B"],
    ],
    # Calibration parameters
    [
        ["Calibration parameter 0", "f"],
        ["Calibration parameter 1", "f"],
        ["Calibration parameter 2", "f116x"],
    ],
]


def read(filename, verbose=False):
    """Parse the binary SPC file and return a dictionary of data.

    ORTEC's SPC file format is divided into records of 128 bytes each. The
    specifications for what each record should contain can be found on pages
    29--44 of this document:
        http://www.ortec-online.com/download/ortec-software-file-structure-manual.pdf

    In the example file, not all of the records that are supposed to be in
    the files seem to be there, and so this code may depart in some places
    from the specification.

    Parameters
    ----------
    filename : str
        The filename of the CNF file to read.
    verbose : bool (optional)
        Whether to print out debugging information. By default False.

    Returns
    -------
    data : dict
        Dictionary of data that can be used to instantiate a Spectrum.
    cal : Calibration
        Energy calibration stored in the file.
    """
    print("SpcFile: Reading file " + filename)
    _, ext = os.path.splitext(filename)
    if ext.lower() != ".spc":
        raise BecquerelParserError("File extension is incorrect: " + ext)

    # initialize a dictionary of spectrum data to populate as we parse
    data = {}

    with open(filename, "rb") as f:
        # read the file in chunks of 128 bytes
        data_records = []
        binary_data = None
        while True:
            if binary_data is not None:
                data_records.append(binary_data)
            try:
                binary_data = f.read(128)
            except OSError:
                raise BecquerelParserError("Unable to read 128 bytes from file")
            if len(binary_data) < 128:
                break
        if verbose:
            print("Done reading in SPC file.  Number of records: ", len(data_records))
        if len(data_records) not in (279, 280):
            raise BecquerelParserError(
                f"Number of data records incorrect: {len(data_records)}"
            )

        # read record data
        i_rec = 0
        for record_format in SPC_FORMAT_BEGINNING:
            if not (
                len(data_records) == 279
                and record_format[0][0] == "Location Description Record 2"
            ):
                binary_data = data_records[i_rec]
                i_rec += 1
                fmt = "<"
                for data_format in record_format:
                    fmt += data_format[1]
                if verbose:
                    print("")
                    print("")
                    print("-" * 60)
                    print("")
                    print(record_format)
                    print(fmt)
                    print("")
                binary_data = struct.unpack(fmt, binary_data)
                if verbose:
                    print("")
                    print(binary_data)
                    print("")
                for j, data_format in enumerate(record_format):
                    if isinstance(binary_data[j], bytes):
                        data[data_format[0]] = binary_data[j].decode("ascii")
                    else:
                        data[data_format[0]] = binary_data[j]
                    if verbose:
                        print(data_format[0], ": ", data[data_format[0]])

        # read spectrum records
        # These records are the spectrum data stored as INTEGER*4
        # numbers beginning with the channel number given and going
        # through the number of channels in the file. They are stored
        # as 64-word records, which gives 32 data channels per record.
        # They are stored sequentially, beginning with the record
        # pointer given.
        i_channel = 0
        channels = []
        counts = []
        for j in range(256):
            binary_data = data_records[i_rec]
            i_rec += 1
            N = struct.unpack("<32I", binary_data)
            # print(': ', N)
            for j, N_j in enumerate(N):
                channels = np.append(channels, i_channel)
                counts = np.append(counts, N_j)
                i_channel += 1

        # read record data
        for record_format in SPC_FORMAT_END:
            binary_data = data_records[i_rec]
            i_rec += 1
            fmt = "<"
            for data_format in record_format:
                fmt += data_format[1]
            if verbose:
                print("")
                print("")
                print("-" * 60)
                print("")
                print(record_format)
                print(fmt)
                print("")
            binary_data = struct.unpack(fmt, binary_data)
            if verbose:
                print("")
                print(binary_data)
                print("")
            for j, data_format in enumerate(record_format):
                if isinstance(binary_data[j], bytes):
                    data[data_format[0]] = binary_data[j].decode("ascii")
                else:
                    data[data_format[0]] = binary_data[j]
                if verbose:
                    print(data_format[0], ": ", data[data_format[0]])

    # finish populating data dict
    data["counts"] = counts

    data["sample_description"] = data["Sample Description"]
    data["detector_description"] = data["Detector Description"]
    if verbose:
        print(data["Start date of sample collection"])
    data["Start date of sample collection"] = data["Start date of sample collection"][
        :-1
    ]
    if verbose:
        print(data["Start date of sample collection"])
        print(data["Start time of sample collection"])
    data["start_time"] = dateutil.parser.parse(
        data["Start date of sample collection"]
        + " "
        + data["Start time of sample collection"]
    )
    if verbose:
        print(data["start_time"])
        print(data["Stop date of sample collection"])
    data["Stop date of sample collection"] = data["Stop date of sample collection"][:-1]
    if verbose:
        print(data["Stop date of sample collection"])
        print(data["Stop time of sample collection"])
    data["collection_stop"] = dateutil.parser.parse(
        data["Stop date of sample collection"]
        + " "
        + data["Stop time of sample collection"]
    )
    data["collection_stop"] = f"{data['collection_stop']:%Y-%m-%dT%H:%M:%S.%f%z}"
    if verbose:
        print(data["collection_stop"])
    data["location_description"] = data["Location Description Record 1"][3:]
    data["location_description"] = (
        data["location_description"].split("\x00\x00\x00")[0].replace("\x00", "\n")
    )
    if len(data_records) > 279:
        data["location_description"] += (
            data["Location Description Record 2"]
            .split("\x00\x00\x00")[0]
            .replace("\x00", "\n")
        )
    data["hardware_status"] = (
        (data["Hardware Parameters Record 1"] + data["Hardware Parameters Record 2"])
        .split("\x00\x00\x00")[0]
        .replace("\x00", "\n")
    )
    data["livetime"] = float(data["Live Time"])
    data["realtime"] = float(data["Real Time"])
    if data["realtime"] <= 0.0:
        raise BecquerelParserError(
            "Realtime not parsed correctly: {}".format(data["realtime"])
        )
    if data["livetime"] <= 0.0:
        raise BecquerelParserError(
            "Livetime not parsed correctly: {}".format(data["livetime"])
        )
    if data["livetime"] > data["realtime"]:
        raise BecquerelParserError(
            "Livetime > realtime: {} > {}".format(data["livetime"], data["realtime"])
        )
    try:
        cal_coeff = [
            float(data["Calibration parameter 0"]),
            float(data["Calibration parameter 1"]),
            float(data["Calibration parameter 2"]),
        ]
    except KeyError:
        raise BecquerelParserError("Calibration parameters not found")

    # clean up null characters in any strings
    for key in data.keys():
        if isinstance(data[key], str):
            data[key] = data[key].replace("\x00", " ")
            data[key] = data[key].replace("\x01", " ")
            data[key] = data[key].strip()

    # create an energy calibration object
    cal = calibration.Calibration.from_polynomial(cal_coeff)

    return data, cal



"""Query photon cross section data from NIST XCOM database.

Provides the fetch_xcom_data function to query the NIST XCOM database,
as well as a few mixtures for common materials.

References:
  https://www.nist.gov/pml/xcom-photon-cross-sections-database
  http://physics.nist.gov/PhysRefData/Xcom/html/xcom1.html

"""

import requests
import pandas as pd
from collections.abc import Iterable
from . import element

# Dry air relative weights taken from:
# http://www.engineeringtoolbox.com/air-composition-d_212.html
MIXTURE_AIR_DRY = [
    "N2 75.47",
    "O2 23.20",
    "Ar 0.933",
    "CO2 0.03",
]

# Seawater composition for 3.5% salinity taken from:
# https://en.wikipedia.org/wiki/Seawater#Compositional_differences_from_freshwater
MIXTURE_SEAWATER = [
    "O 85.84",
    "H 10.82",
    "Cl 1.94",
    "Na 1.08",
    "Mg 0.1292",
    "S 0.091",
    "Ca 0.04",
    "K 0.04",
    "Br 0.0067",
    "C 0.0028",
]

# typical Portland cement values taken from:
# http://matse1.matse.illinois.edu/concrete/prin.html
MIXTURE_PORTLAND_CEMENT = [
    "Ca3SiO5 50",
    "Ca2SiO4 25",
    "Ca3Al2O6 10",
    "Ca4Al2Fe2O10 10",
    "CaSO6H4 5",
]


# base URL for three scripts on NIST where data can be queried
_URL = "https://physics.nist.gov/cgi-bin/Xcom/xcom3_"


# Dictionary of data that will be posted to URL
_DATA = {
    "NumAdd": "1",  # always '1'?
    "Energies": "",  # additional energies separated by ; (MeV)
    "WindowXmin": "0.001",  # lower limit of energy grid (MeV)
    "WindowXmax": "100",  # upper limit of energy grid (MeV)
    "Output": "",  # 'on' for standard energy grid
    "OutOpt": "PIC",  # return cross sections in cm^2/g
    "ResizeFlag": "on",  # seems to determine whether Xmin and Xmax are used
}

# abbreviated names for the table columns (these are used in the dataframe)
COLUMNS_SHORT = [
    "energy",
    "coherent",
    "incoherent",
    "photoelec",
    "pair_nuc",
    "pair_elec",
    "total_w_coh",
    "total_wo_coh",
]

# medium length names for the table columns
COLUMNS_MEDIUM = {
    "energy": "Energy",
    "coherent": "Coherent",
    "incoherent": "Incoherent",
    "photoelec": "Photoelectric",
    "pair_nuc": "Pair production (nuclear)",
    "pair_elec": "Pair Production (electron)",
    "total_w_coh": "Total with coherent",
    "total_wo_coh": "Total without coherent",
}

# verbose names for the table columns
COLUMNS_LONG = {
    "energy": "Photon Energy",
    "coherent": "Coherent Scattering",
    "incoherent": "Incoherent Scattering",
    "photoelec": "Photoelectric Absorption",
    "pair_nuc": "Pair Production in Nuclear Field",
    "pair_elec": "Pair Production in Electron Field",
    "total_w_coh": "Total Attenuation with Coherent Scattering",
    "total_wo_coh": "Total Attenuation without Coherent Scattering",
}


class XCOMError(Exception):
    """General XCOM error."""

    pass


class XCOMInputError(XCOMError):
    """Error related to the user input to XCOMQuery."""

    pass


class XCOMRequestError(XCOMError):
    """Error related to communicating with XCOM or parsing the result."""

    pass


class _XCOMQuery:
    """Query photon cross section data from NIST XCOM database.

    After the data have been successfully queried, they are stored in a
    pandas DataFrame that is accessible through this class's methods.

    The DataFrame column names amd units are:
      'energy': Photon Energy [keV]
      'coherent': coherent scattering cross section [cm^2/g]
      'incoherent': incoherent scattering cross section [cm^2/g]
      'photoelec': photoelectric absorption cross section [cm^2/g]
      'pair_nuc': pair production cross section in nuclear field [cm^2/g]
      'pair_elec': pair production cross section in electron field [cm^2/g]
      'total_w_coh': total attenuation with coherent scattering [cm^2/g]
      'total_wo_coh': total attenuation without coherent scattering [cm^2/g]

    Methods:
      update: add or change the search criteria
      perform: perform the query and parse the results
      __getitem__: use [] to get the DataFrame column
      __len__: len() returns the length of the DataFrame
      __str__: str() returns the str method of the DataFrame
      __format__: format() uses the format method of the DataFrame
      keys: returns the DataFrame keys

    """

    def __init__(self, arg, **kwargs):
        """Initialize and perform an XCOM query.

        Args:
          arg: the atomic number, element symbol, compound string, or mixture
            (the type of argument will be inferred from its content).
          e_range_kev (optional): a length-2 iterable giving the lower and
            upper bounds for a standard grid of energies in keV. Limits must be
            between 1 keV and 1E8 keV, inclusive.
          energies_kev (optional): an iterable of specific energies in keV at
            which cross sections will be evaluated. Energies must be between
            1 keV and 1E8 keV, inclusive.
          perform (optional): set to False to prevent query from immediately
            being performed. [default: True]

        Raises:
          XCOMInputError: if bad search criteria are set or search criteria
            are incomplete.
          XCOMRequestError: if there is a problem with the URL request, or
            a problem parsing the data.

        """
        self._url = _URL
        self._req = None
        self._text = None
        self.df = None
        self._data = dict(_DATA)
        self._method = ""
        # determine which kind of argument 'arg' is (symbol, Z, compound, mix)
        kwargs.update(_XCOMQuery._argument_type(arg))
        self.update(**kwargs)
        if kwargs.get("perform", True):
            self.perform()

    def __len__(self):
        """Pass-through to use DataFrame len()."""
        if self.df is None:
            return 0
        elif len(self.df.keys()) == 0:
            return 0
        else:
            return len(self.df[self.df.keys()[0]])

    def keys(self):
        """Pass-through for DataFrame keys method."""
        return self.df.keys()

    def __getitem__(self, key):
        """Pass-through so that [] accesses the DataFrame."""
        return self.df[key]

    def __str__(self):
        """Pass-through to use DataFrame str method."""
        return str(self.df)

    def __format__(self, formatstr):
        """Pass-through to use DataFrame format method."""
        return self.df.__format__(formatstr)

    @staticmethod
    def _argument_type(arg):
        """Determine if argument is a symbol, Z, compound, or mixture."""
        if isinstance(arg, str):
            if arg.isdigit():
                return {"z": arg}
            elif arg.lower() in [s.lower() for s in element.SYMBOLS]:
                return {"symbol": arg}
            else:
                return {"compound": arg}
        elif isinstance(arg, int):
            return {"z": arg}
        elif isinstance(arg, Iterable):
            return {"mixture": arg}
        raise XCOMInputError(
            f"Cannot determine if argument {arg}"
            + " is a symbol, Z, compound, or mixture"
        )

    @staticmethod
    def _check_z(zstr):
        """Check whether the Z is valid. Raise XCOMInputError if not."""
        zint = int(zstr)
        if zint < 1 or zint > 100:
            raise XCOMInputError(f"XCOM only supports Z from 1 to 100 (z={zstr})")

    @staticmethod
    def _check_compound(formula):
        """Check whether the compound is valid. Raise XCOMInputError if not."""
        if not formula.isalnum():
            raise XCOMInputError(f"Formula not valid: {formula}")

    @staticmethod
    def _check_mixture(formulae):
        """Check whether the mixture is valid. Raise XCOMInputError if not."""
        if not isinstance(formulae, Iterable):
            raise XCOMInputError(f"Mixture formulae must be an iterable: {formulae}")
        for formula in formulae:
            try:
                compound, weight = formula.split()
            except AttributeError:
                raise XCOMInputError(
                    'Mixture formulae "{}" line "{}" must be a string'.format(
                        formulae, formula
                    )
                )
            except ValueError:
                raise XCOMInputError(
                    'Mixture formulae "{}" line "{}" must split into 2'.format(
                        formulae, formula
                    )
                )
            _XCOMQuery._check_compound(compound)
            try:
                float(weight)
            except (ValueError, TypeError):
                raise XCOMInputError(
                    f'Mixture formulae "{formulae}" has bad weight "{weight}"'
                )

    def update(self, **kwargs):
        """Update the search criteria.

        Before calling perform(), one of the following search criteria must
        be set here: z, symbol, compound, or mixture. A valid query will also
        require setting either e_range_kev and/or energies_kev.

        Args:
          symbol (optional): a string of the element symbol, e.g., 'Ge'.
          z (optional): an integer of the element atomic number, e.g., 32.
          compound (optional): a string of the chemical formula, e.g., 'H2O'.
          mixture (optional): a list of compounds and relative weights, e.g.,
            ['H2O 0.5', 'Ge 0.5']
          e_range_kev (optional): a length-2 iterable giving the lower and
            upper bounds for a standard grid of energies in keV. Limits must be
            between 1 keV and 1E8 keV, inclusive.
          energies_kev (optional): an iterable of specific energies in keV at
            which cross sections will be evaluated. Energies must be between
            1 keV and 1E8 keV, inclusive.

        Raises:
          XCOMInputError: if bad search criteria are set.

        """
        # check for valid keywords
        for kwarg in kwargs:
            if kwarg not in [
                "symbol",
                "z",
                "compound",
                "mixture",
                "e_range_kev",
                "energies_kev",
                "perform",
            ]:
                raise XCOMInputError(f'Unknown keyword: "{kwarg}"')

        # determine the search method (element, compound, or mixture)
        if "symbol" in kwargs:
            self._method = "1"
            sym = kwargs["symbol"]
            self._data["ZSym"] = sym
        elif "z" in kwargs:
            self._method = "1"
            znum = kwargs["z"]
            _XCOMQuery._check_z(znum)
            self._data["ZNum"] = f"{int(znum):d}"
        elif "compound" in kwargs:
            # convert compound to mixture to avoid occasional problem
            # with XCOM compound queries (see issue #76)
            # self._method = '2'
            self._method = "3"
            formula = kwargs["compound"]
            _XCOMQuery._check_compound(formula)
            formulae = [formula + " 1"]
            _XCOMQuery._check_mixture(formulae)
            formulae = "\r\n".join(formulae)
            # self._data['Formula'] = formula
            self._data["Formulae"] = formulae
        elif "mixture" in kwargs:
            self._method = "3"
            formulae = kwargs["mixture"]
            _XCOMQuery._check_mixture(formulae)
            formulae = "\r\n".join(formulae)
            self._data["Formulae"] = formulae

        # include standard grid of energies
        if "e_range_kev" in kwargs:
            if not isinstance(kwargs["e_range_kev"], Iterable):
                raise XCOMInputError(
                    "XCOM e_range_kev must be iterable of length 2: {}".format(
                        kwargs["e_range_kev"]
                    )
                )
            if len(kwargs["e_range_kev"]) != 2:
                raise XCOMInputError(
                    "XCOM e_range_kev must be iterable of length 2: {}".format(
                        kwargs["e_range_kev"]
                    )
                )
            if kwargs["e_range_kev"][0] < 1:
                raise XCOMInputError(
                    "XCOM e_range_kev[0] must be >= 1 keV: {}".format(
                        kwargs["e_range_kev"][0]
                    )
                )
            if kwargs["e_range_kev"][1] > 1e8:
                raise XCOMInputError(
                    "XCOM e_range_kev[1] must be <= 1E8 keV: {}".format(
                        kwargs["e_range_kev"][1]
                    )
                )
            if kwargs["e_range_kev"][0] >= kwargs["e_range_kev"][1]:
                raise XCOMInputError(
                    "XCOM e_range_kev[0] must be < e_range_kev[1]: {}".format(
                        kwargs["e_range_kev"]
                    )
                )
            self._data["WindowXmin"] = "{:.6f}".format(
                kwargs["e_range_kev"][0] / 1000.0
            )
            self._data["WindowXmax"] = "{:.6f}".format(
                kwargs["e_range_kev"][1] / 1000.0
            )
            self._data["Output"] = "on"

        # additional energies
        if "energies_kev" in kwargs:
            if not isinstance(kwargs["energies_kev"], Iterable):
                raise XCOMInputError(
                    "XCOM energies_kev must be an iterable: {}".format(
                        kwargs["energies_kev"]
                    )
                )
            for energy in kwargs["energies_kev"]:
                if energy < 1 or energy > 1e8:
                    raise XCOMInputError(
                        f"XCOM energy must be >= 1 and <= 1E8 keV: {energy}"
                    )
            self._data["Energies"] = ";".join(
                [f"{erg / 1000.0:.6f}" for erg in kwargs["energies_kev"]]
            )

    def _request(self):
        """Request data table from the URL."""
        self._req = requests.post(self._url + self._method, data=self._data)
        if not self._req.ok or self._req.reason != "OK" or self._req.status_code != 200:
            raise XCOMRequestError(
                "XCOM Request failed: reason={}, status_code={}".format(
                    self._req.reason, self._req.status_code
                )
            )
        if "Error" in self._req.text:
            raise XCOMRequestError(f"XCOM returned an error:\n{self._req.text}")

    def _parse_text(self):
        """Parse table contained in the text into a dictionary."""
        self._text = str(self._req.text)
        if len(self._text) == 0:
            raise XCOMRequestError("XCOM returned no text")
        tables = pd.read_html(self._text, header=0, skiprows=[1, 2])
        if len(tables) != 1:
            raise XCOMRequestError("More than one HTML table found")
        self.df = tables[0]
        if len(self.df.keys()) != 1 + len(COLUMNS_SHORT):
            raise XCOMRequestError(
                "Found {} columns but expected {}".format(
                    len(self.df.keys()), 1 + len(COLUMNS_SHORT)
                )
            )
        # remove 'edge' column
        self.df = self.df[self.df.keys()[1:]]
        self.df.columns = COLUMNS_SHORT
        if len(self) == 0:
            raise XCOMRequestError("Parsed DataFrame is empty")

    def perform(self):
        """Perform the query.

        Before calling perform(), set the search criteria using update()
        or __init__().

        Raises:
          XCOMInputError: if search criteria are incomplete.
          XCOMRequestError: if there is a problem with the URL request, or
            a problem parsing the data.

        """
        if self._method not in ["1", "2", "3"]:
            raise XCOMInputError(
                "XCOM search method not set. Need to call update() method."
            )
        if self._data["Energies"] == "" and self._data["Output"] == "":
            raise XCOMInputError("No energies_kev or e_range_kev requested.")
        # submit the query
        self._request()
        # package the output into a pandas DataFrame
        self._parse_text()
        # convert energy from MeV to keV
        self.df["energy"] *= 1000.0


def fetch_xcom_data(arg, **kwargs):
    """Query photon cross section data from NIST XCOM database.

    Returns a pandas DataFrame containing the cross section data.

    The DataFrame column names amd units are:
      'energy': Photon Energy [keV]
      'coherent': coherent scattering cross section [cm^2/g]
      'incoherent': incoherent scattering cross section [cm^2/g]
      'photoelec': photoelectric absorption cross section [cm^2/g]
      'pair_nuc': pair production cross section in nuclear field [cm^2/g]
      'pair_elec': pair production cross section in electron field [cm^2/g]
      'total_w_coh': total attenuation with coherent scattering [cm^2/g]
      'total_wo_coh': total attenuation without coherent scattering [cm^2/g]

    Args:
      arg: the atomic number, element symbol, compound string, or mixture
        (the type of argument will be inferred from its content).
      e_range_kev (optional): a length-2 iterable giving the lower and
        upper bounds for a standard grid of energies in keV. Limits must be
        between 1 keV and 1E8 keV, inclusive.
      energies_kev (optional): an iterable of specific energies in keV at
        which cross sections will be evaluated. Energies must be between
        1 keV and 1E8 keV, inclusive.

    Raises:
      XCOMInputError: if bad search criteria are set or search criteria
        are incomplete.
      XCOMRequestError: if there is a problem with the URL request, or
        a problem parsing the data.

    """
    query = _XCOMQuery(arg, **kwargs)
    return query.df



"""Nuclear isotopes and isomers."""

from future.builtins import super
import numpy as np
import uncertainties
from . import element
from .wallet_cache import wallet_cache

N_AV = 6.022141e23  # mol^-1


class IsotopeError(element.ElementError):
    """Problem with isotope properties."""

    pass


def _split_element_mass(arg):
    """Split a string into an element name/symbol plus a mass number/isomer.

    For example, the string 'TC99M' would be split into ('TC', '99M'), the
    string '238U' would be split into ('U', '238'), and the string 'Hf-178m3'
    would be split into ('Hf', '178m3').

    Args:
      arg: a string of the form "EE[-]AAA[m[M]]" or "AAA[m[M]][-]EE".

    Returns:
      tuple of strings identified as the element symbol/name and mass
        number A, which may include isomer level as well.

    Raises:
      IsotopeError: if the string could not be split.
    """

    arg = str(arg)
    element_id = None
    mass_isomer = None
    if "-" in arg:
        # parse string of the form "EE-AAA[m[M]]" or "AAA[m[M]]-EE"
        # (i.e., those with a hyphen)
        tokens = arg.split("-")
        if len(tokens) != 2:
            raise IsotopeError(f"Too many hyphens in isotope string: {arg}")
        if tokens[0].isalpha() and tokens[1][0].isdigit():
            element_id = tokens[0]
            mass_isomer = tokens[1]
        elif tokens[0][0].isdigit() and tokens[1].isalpha():
            element_id = tokens[1]
            mass_isomer = tokens[0]
        else:
            raise IsotopeError(f"Could not find mass number for isotope: {tokens}")
    else:
        # parse string of the form "EEAAA[m[M]]" or "AAA[m[M]]EE"
        # (i.e., those without a hyphen)
        part1 = [arg[j:] for j in range(1, len(arg))]
        part2 = [arg[:j] for j in range(1, len(arg))]
        element_ids = []
        mass_isomers = []
        # make a list of possible element symbols and names
        for p1, p2 in zip(part1, part2):
            if p1[0].isdigit() and p2[0].isalpha():
                p_element = p2
                p_mass = p1
            elif p1[0].isalpha() and p2[0].isdigit():
                p_element = p1
                p_mass = p2
            else:
                continue
            try:
                element.Element(p_element)
            except element.ElementError:
                continue
            element_ids.append(p_element)
            mass_isomers.append(p_mass)
        if len(element_ids) == 0:
            raise IsotopeError(f"Could not find element for isotope: {arg}")
        # if multiple element IDs were found, choose the longest
        element_id = ""
        mass_isomer = ""
        for elem, mass in zip(element_ids, mass_isomers):
            if len(elem) > len(element_id):
                element_id = elem
                mass_isomer = mass
    # ensure element name or symbol is valid
    try:
        element.Element(element_id)
    except element.ElementError:
        raise IsotopeError(f"Element name or symbol is invalid: {element_id}")
    return element_id, mass_isomer


def _split_mass_isomer(arg):
    """Split a string into a mass number and isomer level.

    For example, the string '99m' would be split into ('99', 'm'), the
    string '238' would be split into ('238', ''), and the string '178m3'
    would be split into ('178', 'm3').

    Args:
      arg: a string of the form "AAA[m[M]]".

    Returns:
      tuple of substrings identified as the mass number A and isomer.

    Raises:
      IsotopeError: if the string could not be split.
    """

    arg = str(arg)
    aa = 0
    mm = ""
    if "m" in arg.lower():
        tokens = arg.lower().split("m")
        if len(tokens) != 2:
            raise IsotopeError(f"Too many ms in mass number: {arg} {tokens}")
        try:
            aa = int(tokens[0])
        except ValueError:
            raise IsotopeError(
                f"Mass number cannot be converted to int: {tokens[0]} {arg}"
            )
        mm = "m"
        if len(tokens[1]) > 0:
            if not tokens[1].isdigit():
                raise IsotopeError(
                    f"Metastable level must be a number: {tokens[1]} {arg}"
                )
            mm += tokens[1]
    else:
        try:
            aa = int(arg)
        except ValueError:
            raise IsotopeError(f"Mass number cannot be converted to int: {arg}")
    return (aa, mm)


def parse_isotope(arg):
    """Parse an isotope string into a symbol, mass, and metastable state.

    Args:
      arg: A string identifying the isotope, such as "232TH", "U-238", or
        "Tc-99m".

    Returns:
      A tuple of the element symbol, mass number, and metastable state.

    Raises:
      IsotopeError: if there was a problem parsing the string.
    """

    element_id, mass_isomer = _split_element_mass(arg)
    mass_number, isomer_level = _split_mass_isomer(mass_isomer)
    return (element_id, mass_number, isomer_level)


class Isotope(element.Element):
    """Basic properties of a nuclear isotope, including isomers.

    Also provides string formatting:
    >>> iso = Isotope('178M2HF')
    >>> '{:%n(%s)-%a%m Z=%z A=%a}'.format(iso)
    'Hafnium(Hf)-178m2 Z=72 A=178'

    Properties (read-only):
      symbol: the mixed-case element symbol (e.g., "Ge")
      name: the mixed-case element name (e.g., "Germanium")
      Z: an integer giving the atomic number (e.g., 32)
      atomic_mass: a float giving the atomic mass in amu
      A: an integer giving the mass number (e.g., 68)
      N: an integer giving the neutron number (e.g., 36)
      M: an integer giving the isomer level (e.g., 0)
      m: string version of M, e.g., '', 'm', or 'm2'
      half_life: half-life of the isotope in seconds
      is_stable: a boolean of whether the isotope is stable
      abundance: the natural abundance percent
      j_pi: a string describing the spin and parity
      energy_level: the nuclear energy level in MeV
      mass_excess: the mass excess in MeV
      decay_modes: the isotope's decay modes and their branching ratios
      specific_activity: the specific activity in Bq/g
    """

    def __init__(self, *args):
        """Initialize an Isotope.

        Args:
          args: Either one argument that is a string describing the isotope,
            e.g., "Tc-99m", or two or three arguments: one of the set
            {`name`, `sym`, `Z`} and the mass number `A`, and optionally the
            isomer level `m`.

        Raises:
          IsotopeError: if there was a problem instantiating the isotope.
        """

        self.A = 0
        self.m = ""
        self.M = 0
        if len(args) == 1:
            if not isinstance(args[0], str):
                raise IsotopeError("Single argument must be a string")
            symbol, aa, mm = parse_isotope(args[0])
            super().__init__(symbol)
            self._init_A(aa)
            self._init_m(mm)
        elif len(args) == 2 or len(args) == 3:
            try:
                super().__init__(args[0])
            except element.ElementError:
                raise IsotopeError(f"Unable to create Isotope: {args}")
            self._init_A(args[1])
            if len(args) == 3:
                self._init_m(args[2])
        else:
            raise IsotopeError("One, two, or three arguments required")
        self.N = self.A - self.Z
        if self.N < 0:
            raise IsotopeError(f"Neutron number N cannot be negative: {args} {self.N}")

    def _init_A(self, arg):
        """Initialize with an isotope A."""
        try:
            self.A = int(arg)
        except ValueError:
            raise IsotopeError(f"Mass number cannot be converted to integer: {arg}")
        if self.A < 1:
            raise IsotopeError(f"Mass number must be >= 1: {self.A}")

    def _init_m(self, arg):
        """Initialize with an isomer level number."""
        if arg == "" or arg is None or arg == 0:
            self.m = ""
            self.M = 0
        else:
            if isinstance(arg, int):
                if arg == 1:
                    self.m = "m"
                    self.M = 1
                elif arg >= 2:
                    self.M = arg
                    self.m = f"m{self.M}"
                else:
                    raise IsotopeError(f"Metastable level must be >= 0: {arg}")
            elif isinstance(arg, str):
                self.m = arg.lower()
                if self.m[0] != "m":
                    raise IsotopeError(
                        f'Metastable level must start with "m": {self.m}'
                    )
                if len(self.m) > 1:
                    if not self.m[1:].isdigit():
                        raise IsotopeError(
                            "Metastable level must be numeric: {} {}".format(
                                self.m[0], self.m[1:]
                            )
                        )
                    self.M = int(self.m[1:])
                else:
                    self.M = 1
            else:
                raise IsotopeError(
                    "Metastable level must be integer or string: {} {}".format(
                        arg, type(arg)
                    )
                )

    def __str__(self):
        """Define behavior of str() on Isotope."""
        return f"{self}"

    def __format__(self, formatstr):
        """Define behavior of string's format method.

        Format codes:
            '%s': element symbol
            '%n': element name
            '%z': element Z
            '%a': isotope A
            '%m': isomer level
        """

        str0 = str(formatstr)
        if len(str0) == 0:
            str0 = "%s-%a%m"
        str0 = str0.replace("%s", self.symbol)
        str0 = str0.replace("%n", self.name)
        str0 = str0.replace("%z", f"{self.Z}")
        str0 = str0.replace("%a", f"{self.A}")
        str0 = str0.replace("%m", self.m)
        return str0

    def __eq__(self, other):
        """Define equality of two isotopes."""
        if isinstance(other, Isotope):
            return (
                super().__eq__(other)
                and self.A == other.A
                and self.Z == other.Z
                and self.M == other.M
            )
        else:
            raise TypeError("Cannot compare to non-isotope")

    def _wallet_card(self):
        """Retrieve the wallet card data for this isotope.

        Returns:
          a DataFrame containing the wallet card data.

        Raises:
          IsotopeError: if no isotope data can be found.
        """

        if not wallet_cache.loaded:
            wallet_cache.load()
        this_isotope = (
            (wallet_cache.df["Z"] == self.Z)
            & (wallet_cache.df["A"] == self.A)
            & (wallet_cache.df["M"] == self.M)
        )
        df = wallet_cache.df[this_isotope]
        if len(df) == 0:
            raise IsotopeError(f"No wallet card data found for isotope {self}")
        return df

    @property
    def half_life(self):
        """The isotope's half-life in seconds.

        Returns:
          the half-life in seconds, or np.inf if stable.
        """

        df = self._wallet_card()
        data = df["T1/2 (s)"].tolist()
        assert len(np.unique(data)) == 1
        return data[0]

    @property
    def decay_const(self):
        """The decay constant (lambda), in 1/seconds.

        Returns:
          the decay constant in 1/seconds, or 0 if stable.
        """

        return np.log(2) / self.half_life

    @property
    def is_stable(self):
        """Return a boolean of whether the isotope is stable.

        Returns:
          a boolean.
        """

        df = self._wallet_card()
        data = df["T1/2 (txt)"].tolist()
        assert len(np.unique(data)) == 1
        return "STABLE" in data

    @property
    def abundance(self):
        """Return the natural abundance percent.

        Returns:
          the natural abundance (float or ufloat), or None if not stable.
        """

        df = self._wallet_card()
        data = df["Abundance (%)"].tolist()
        if not isinstance(data[0], uncertainties.core.Variable):
            if np.isnan(data[0]):
                return None
        return data[0]

    @property
    def j_pi(self):
        """Return the spin and parity of the isotope.

        Returns:
          a string giving the J and Pi, e.g., '7/2+'.
        """

        df = self._wallet_card()
        data = df["JPi"].tolist()
        assert len(np.unique(data)) == 1
        return data[0]

    @property
    def energy_level(self):
        """Return the nuclear energy level in MeV.

        Returns:
          the energy level in MeV.
        """

        df = self._wallet_card()
        data = df["Energy Level (MeV)"].tolist()
        assert len(np.unique(data)) == 1
        return data[0]

    @property
    def mass_excess(self):
        """Return the mass excess in MeV.

        Returns:
          the mass excess in MeV.
        """

        df = self._wallet_card()
        data = df["Mass Excess (MeV)"].tolist()
        if not isinstance(data[0], uncertainties.core.Variable):
            if np.isnan(data[0]):
                return None
        return data[0]

    @property
    def decay_modes(self):
        """Return the decay modes and their branching ratios.

        Returns:
          a list of the decay modes, and a list of the branching percents.
        """

        df = self._wallet_card()
        data1 = df["Decay Mode"].tolist()
        data2 = df["Branching (%)"].tolist()
        if len(data1) == 1 and np.isnan(data2[0]):
            return [], []
        return data1, data2

    @property
    def specific_activity(self):
        """Return the specific activity in Bq/g."""

        return self.decay_const * N_AV / self.A



"""Simple tools to perform HDF5 I/O."""

import pathlib
from typing import Union, Tuple
import h5py


def ensure_string(data):
    """Ensure the data are decoded to a string if they are bytes.

    Parameters
    ----------
    data : str or bytes
    """
    if isinstance(data, str):
        return data
    elif isinstance(data, bytes):
        return data.decode("ascii", "replace")
    else:
        raise TypeError(f"Data are neither bytes nor string: {type(data)}")


def is_h5_filename(name: str):
    """Return True if the lowercase filename ends with h5 or hdf5.

    Parameters
    ----------
    name : str
        The filename to examine.

    Returns
    -------
    is_h5 : bool
    """
    try:
        ext = pathlib.Path(name).suffix
        return ext.lower().endswith((".h5", ".hdf5"))
    except TypeError:
        return False


class open_h5:
    """Context manager to allow I/O given HDF5 filename, File, or Group."""

    def __init__(self, name: Union[str, h5py.File, h5py.Group], mode=None, **kwargs):
        """Initialize the context manager.

        Parameters
        ----------
        name : str, h5py.File, h5py.Group
            The filename or an open h5py File or Group.
        mode : str
            The I/O mode. Default is None.
        kwargs : dict
            Additional keyword arguments sent to h5py.File if initializing.
        """
        self._already_h5_obj = isinstance(name, (h5py.File, h5py.Group))
        if not self._already_h5_obj:
            assert is_h5_filename(name), "Filename must end in h5 or hdf5"
        self.name = name
        self.mode = mode
        self.kwargs = kwargs
        self.file = None

    def __enter__(self):
        """Open the file for I/O.

        Returns
        -------
        file : h5py.File, h5py.Group
            An open h5py File or Group.
        """
        if self._already_h5_obj:
            self.file = self.name
        else:
            self.file = h5py.File(self.name, mode=self.mode, **self.kwargs)
        return self.file

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close the file if it had to open it in __enter__."""
        if not self._already_h5_obj:
            self.file.close()


def write_h5(name: Union[str, h5py.File, h5py.Group], dsets: dict, attrs: dict) -> None:
    """Write the datasets and attributes to an HDF5 file or group.

    Parameters
    ----------
    name : str, h5py.File, h5py.Group
        The filename or an open h5py File or Group.
    dsets : dict
        Dictionary of data to be written as datasets.
    attrs : dict
        Dictionary of data to be written as attributes.
    """
    with open_h5(name, "w") as file:
        # write the datasets
        for key in dsets.keys():
            try:
                file.create_dataset(
                    key,
                    data=dsets[key],
                    compression="gzip",
                    compression_opts=9,
                )
            except TypeError:
                file.create_dataset(key, data=dsets[key])
        # write the attributes
        file.attrs.update(attrs)


def read_h5(name: Union[str, h5py.File, h5py.Group]) -> Tuple[dict, dict, list]:
    """Read the datasets and attributes from an HDF5 file or group.

    Parameters
    ----------
    name : str, h5py.File, h5py.Group
        The filename or an open h5py File or Group.

    Returns
    -------
    dsets : dict
        Dictionary of data to be written as datasets.
    attrs : dict
        Dictionary of data to be written as attributes.
    skipped : list
        List of keys of skipped items.
    """
    dsets = {}
    attrs = {}
    skipped = []
    with open_h5(name, "r") as file:
        # read the datasets
        for key in file.keys():
            # skip any non-datasets
            if not isinstance(file[key], h5py.Dataset):
                skipped.append(str(key))
            else:
                dsets[key] = file[key][()]
        # read the attributes
        attrs = dict(file.attrs)
    return dsets, attrs, skipped



"""Tools for radiation spectral analysis."""

from . import xcom, nndc, materials, element, isotope, isotope_qty
from .element import (
    Element,
    ElementError,
    ElementZError,
    ElementSymbolError,
    ElementNameError,
)
from .isotope import Isotope, IsotopeError
from .isotope_qty import (
    IsotopeQuantity,
    IsotopeQuantityError,
    NeutronIrradiation,
    NeutronIrradiationError,
    UCI_TO_BQ,
    N_AV,
)
from .materials import (
    fetch_element_data,
    fetch_compound_data,
    NISTMaterialsError,
    NISTMaterialsRequestError,
)
from .nndc import (
    fetch_wallet_card,
    fetch_decay_radiation,
    NNDCError,
    NoDataFound,
    NNDCInputError,
    NNDCRequestError,
)
from .xcom import (
    fetch_xcom_data,
    XCOMError,
    XCOMInputError,
    XCOMRequestError,
    MIXTURE_AIR_DRY,
    MIXTURE_SEAWATER,
    MIXTURE_PORTLAND_CEMENT,
)

__all__ = [
    "xcom",
    "nndc",
    "materials",
    "element",
    "isotope",
    "isotope_qty",
    "Element",
    "ElementError",
    "ElementZError",
    "ElementSymbolError",
    "ElementNameError",
    "Isotope",
    "IsotopeError",
    "IsotopeQuantity",
    "IsotopeQuantityError",
    "NeutronIrradiation",
    "NeutronIrradiationError",
    "UCI_TO_BQ",
    "N_AV",
    "fetch_element_data",
    "fetch_compound_data",
    "NISTMaterialsError",
    "NISTMaterialsRequestError",
    "fetch_wallet_card",
    "fetch_decay_radiation",
    "NNDCError",
    "NoDataFound",
    "NNDCInputError",
    "NNDCRequestError",
    "fetch_xcom_data",
    "XCOMError",
    "XCOMInputError",
    "XCOMRequestError",
    "MIXTURE_AIR_DRY",
    "MIXTURE_SEAWATER",
    "MIXTURE_PORTLAND_CEMENT",
]



"""Code for parsing spectrum file types."""

from .parsers import BecquerelParserError, BecquerelParserWarning
from . import h5, cnf, spc, spe

__all__ = ["BecquerelParserError", "BecquerelParserWarning", "h5", "cnf", "spc", "spe"]



"""Define errors and warnings for I/O."""

import warnings

warnings.simplefilter("always", DeprecationWarning)


class BecquerelParserWarning(UserWarning):
    """Warnings encountered during parsing."""

    pass


class BecquerelParserError(Exception):
    """Failure encountered while parsing."""

    pass



"""Element symbols, names, Z values, and masses."""


# Element data (Z, symbol, name, and relative atomic mass) are from Wikipedia:
# https://en.wikipedia.org/wiki/List_of_chemical_elements

_Z_SYMBOL_NAME_MASS = (
    (1, "H", "Hydrogen", 1.01),
    (2, "He", "Helium", 4.00),
    (3, "Li", "Lithium", 6.94),
    (4, "Be", "Beryllium", 9.01),
    (5, "B", "Boron", 10.81),
    (6, "C", "Carbon", 12.01),
    (7, "N", "Nitrogen", 14.01),
    (8, "O", "Oxygen", 16.00),
    (9, "F", "Fluorine", 19.00),
    (10, "Ne", "Neon", 20.18),
    (11, "Na", "Sodium", 22.99),
    (12, "Mg", "Magnesium", 24.30),
    (13, "Al", "Aluminum", 26.98),
    (14, "Si", "Silicon", 28.09),
    (15, "P", "Phosphorus", 30.97),
    (16, "S", "Sulfur", 32.06),
    (17, "Cl", "Chlorine", 35.45),
    (18, "Ar", "Argon", 39.95),
    (19, "K", "Potassium", 39.10),
    (20, "Ca", "Calcium", 40.08),
    (21, "Sc", "Scandium", 44.96),
    (22, "Ti", "Titanium", 47.87),
    (23, "V", "Vanadium", 50.94),
    (24, "Cr", "Chromium", 52.00),
    (25, "Mn", "Manganese", 54.94),
    (26, "Fe", "Iron", 55.84),
    (27, "Co", "Cobalt", 58.93),
    (28, "Ni", "Nickel", 58.69),
    (29, "Cu", "Copper", 63.55),
    (30, "Zn", "Zinc", 65.38),
    (31, "Ga", "Gallium", 69.72),
    (32, "Ge", "Germanium", 72.63),
    (33, "As", "Arsenic", 74.92),
    (34, "Se", "Selenium", 78.97),
    (35, "Br", "Bromine", 79.90),
    (36, "Kr", "Krypton", 83.80),
    (37, "Rb", "Rubidium", 85.47),
    (38, "Sr", "Strontium", 87.62),
    (39, "Y", "Yttrium", 88.91),
    (40, "Zr", "Zirconium", 91.22),
    (41, "Nb", "Niobium", 92.91),
    (42, "Mo", "Molybdenum", 95.95),
    (43, "Tc", "Technetium", 98),
    (44, "Ru", "Ruthenium", 101.07),
    (45, "Rh", "Rhodium", 102.91),
    (46, "Pd", "Palladium", 106.42),
    (47, "Ag", "Silver", 107.87),
    (48, "Cd", "Cadmium", 112.41),
    (49, "In", "Indium", 114.82),
    (50, "Sn", "Tin", 118.71),
    (51, "Sb", "Antimony", 121.76),
    (52, "Te", "Tellurium", 127.60),
    (53, "I", "Iodine", 126.90),
    (54, "Xe", "Xenon", 131.29),
    (55, "Cs", "Cesium", 132.91),
    (56, "Ba", "Barium", 137.33),
    (57, "La", "Lanthanum", 138.91),
    (58, "Ce", "Cerium", 140.12),
    (59, "Pr", "Praseodymium", 140.91),
    (60, "Nd", "Neodymium", 144.24),
    (61, "Pm", "Promethium", 145),
    (62, "Sm", "Samarium", 150.36),
    (63, "Eu", "Europium", 151.96),
    (64, "Gd", "Gadolinium", 157.25),
    (65, "Tb", "Terbium", 158.93),
    (66, "Dy", "Dysprosium", 162.50),
    (67, "Ho", "Holmium", 164.93),
    (68, "Er", "Erbium", 167.26),
    (69, "Tm", "Thulium", 168.93),
    (70, "Yb", "Ytterbium", 173.04),
    (71, "Lu", "Lutetium", 174.97),
    (72, "Hf", "Hafnium", 178.49),
    (73, "Ta", "Tantalum", 180.95),
    (74, "W", "Tungsten", 183.84),
    (75, "Re", "Rhenium", 186.21),
    (76, "Os", "Osmium", 190.23),
    (77, "Ir", "Iridium", 192.22),
    (78, "Pt", "Platinum", 195.08),
    (79, "Au", "Gold", 196.97),
    (80, "Hg", "Mercury", 200.59),
    (81, "Tl", "Thallium", 204.38),
    (82, "Pb", "Lead", 207.20),
    (83, "Bi", "Bismuth", 208.98),
    (84, "Po", "Polonium", 209),
    (85, "At", "Astatine", 210),
    (86, "Rn", "Radon", 222),
    (87, "Fr", "Francium", 223),
    (88, "Ra", "Radium", 226),
    (89, "Ac", "Actinium", 227),
    (90, "Th", "Thorium", 232.04),
    (91, "Pa", "Protactinium", 231.04),
    (92, "U", "Uranium", 238.03),
    (93, "Np", "Neptunium", 237),
    (94, "Pu", "Plutonium", 244),
    (95, "Am", "Americium", 243),
    (96, "Cm", "Curium", 247),
    (97, "Bk", "Berkelium", 247),
    (98, "Cf", "Californium", 251),
    (99, "Es", "Einsteinium", 252),
    (100, "Fm", "Fermium", 257),
    (101, "Md", "Mendelevium", 258),
    (102, "No", "Nobelium", 259),
    (103, "Lr", "Lawrencium", 266),
    (104, "Rf", "Rutherfordium", 267),
    (105, "Db", "Dubnium", 268),
    (106, "Sg", "Seaborgium", 269),
    (107, "Bh", "Bohrium", 270),
    (108, "Hs", "Hassium", 269),
    (109, "Mt", "Meitnerium", 278),
    (110, "Ds", "Darmstadtium", 281),
    (111, "Rg", "Roentgenium", 282),
    (112, "Cn", "Copernicium", 285),
    (113, "Nh", "Nihonium", 286),
    (114, "Fl", "Flerovium", 289),
    (115, "Mc", "Moscovium", 289),
    (116, "Lv", "Livermorium", 293),
    (117, "Ts", "Tennessine", 294),
    (118, "Og", "Oganesson", 294),
)


ZS = {d[0] for d in _Z_SYMBOL_NAME_MASS}
SYMBOLS = {d[1] for d in _Z_SYMBOL_NAME_MASS}
SYMBOLS_LOWER = {d[1].lower() for d in _Z_SYMBOL_NAME_MASS}
SYMBOLS_UPPER = {d[1].upper() for d in _Z_SYMBOL_NAME_MASS}
NAMES = {d[2] for d in _Z_SYMBOL_NAME_MASS}
NAMES_LOWER = {d[2].lower() for d in _Z_SYMBOL_NAME_MASS}
NAMES_UPPER = {d[2].upper() for d in _Z_SYMBOL_NAME_MASS}

_SYMBOL_FROM_SYMBOL_LOWER = {d[1].lower(): d[1] for d in _Z_SYMBOL_NAME_MASS}
_SYMBOL_FROM_SYMBOL_UPPER = {d[1].upper(): d[1] for d in _Z_SYMBOL_NAME_MASS}

_NAME_FROM_NAME_LOWER = {d[2].lower(): d[2] for d in _Z_SYMBOL_NAME_MASS}
_NAME_FROM_NAME_UPPER = {d[2].upper(): d[2] for d in _Z_SYMBOL_NAME_MASS}

_Z_FROM_SYMBOL = {d[1]: d[0] for d in _Z_SYMBOL_NAME_MASS}
_Z_FROM_NAME = {d[2]: d[0] for d in _Z_SYMBOL_NAME_MASS}

_SYMBOL_FROM_Z = {d[0]: d[1] for d in _Z_SYMBOL_NAME_MASS}
_SYMBOL_FROM_NAME = {d[2]: d[1] for d in _Z_SYMBOL_NAME_MASS}

_NAME_FROM_Z = {d[0]: d[2] for d in _Z_SYMBOL_NAME_MASS}
_NAME_FROM_SYMBOL = {d[1]: d[2] for d in _Z_SYMBOL_NAME_MASS}

_MASS_FROM_Z = {d[0]: d[3] for d in _Z_SYMBOL_NAME_MASS}
_MASS_FROM_SYMBOL = {d[1]: d[3] for d in _Z_SYMBOL_NAME_MASS}


class ElementError(Exception):
    """Problem with element properties."""

    pass


class ElementZError(ElementError):
    """Bad element Z value."""

    pass


class ElementSymbolError(ElementError):
    """Bad element symbol."""

    pass


class ElementNameError(ElementError):
    """Bad element name."""

    pass


def validated_z(z):
    """Convert z into a valid element atomic number.

    Args:
      z: numeric or string type representing the atomic number Z.

    Returns:
      Integer z that is a valid atomic number.

    Raises:
      ElementZError: if z cannot be converted to integer or is out of range.
    """

    try:
        int(z)
    except ValueError:
        raise ElementZError(f'Element Z="{z}" invalid')
    if int(z) not in ZS:
        raise ElementZError(f'Element Z="{z}" not found')
    return int(z)


def validated_symbol(sym):
    """Convert symbol into a valid mixed-case element symbol.

    Args:
      sym: a string type

    Returns:
      A valid mixed-case element symbol that matches sym.

    Raises:
      ElementSymbolError: if sym does not match a valid symbol.
    """

    try:
        sym.lower()
    except AttributeError:
        raise ElementSymbolError(f'Element symbol "{sym}" invalid')
    if sym.lower() not in SYMBOLS_LOWER:
        raise ElementSymbolError(f'Element symbol "{sym}" not found')
    return _SYMBOL_FROM_SYMBOL_LOWER[sym.lower()]


def validated_name(nm):
    """Convert name into a valid mixed-case element name.

    Args:
      nm: a string type

    Returns:
      A valid mixed-case element name that matches nm.

    Raises:
      ElementNameError: if nm does not match a valid name.
    """

    try:
        nm.lower()
    except AttributeError:
        raise ElementNameError(f'Element name "{nm}" invalid')
    # special case: Alumin[i]um
    if nm.lower() == "aluminium":
        nm = "Aluminum"
    # special case: C[a]esium
    if nm.lower() == "caesium":
        nm = "Cesium"
    if nm.lower() not in NAMES_LOWER:
        raise ElementNameError(f'Element name "{nm}" not found')
    return _NAME_FROM_NAME_LOWER[nm.lower()]


def element_z(sym_or_name):
    """Convert element symbol or name into a valid element atomic number Z.

    Args:
      sym_or_name: string type representing an element symbol or name.

    Returns:
      Integer z that is a valid atomic number matching the symbol or name.

    Raises:
      ElementZError: if the symbol or name cannot be converted.
    """

    try:
        return _Z_FROM_SYMBOL[validated_symbol(sym_or_name)]
    except ElementSymbolError:
        pass
    try:
        return _Z_FROM_NAME[validated_name(sym_or_name)]
    except ElementNameError:
        raise ElementZError("Must supply either the element symbol or name")


def element_symbol(name_or_z):
    """Convert element name or Z into a valid mixed-case element symbol.

    Args:
      name_or_z: string or numeric type representing an element name or Z.

    Returns:
      A valid mixed-case element symbol that matches the name or Z.

    Raises:
      ElementSymbolError: if the symbol or name cannot be converted.
    """

    try:
        return _SYMBOL_FROM_Z[validated_z(name_or_z)]
    except ElementZError:
        pass
    try:
        return _SYMBOL_FROM_NAME[validated_name(name_or_z)]
    except ElementNameError:
        raise ElementSymbolError("Must supply either the Z or element name")


def element_name(sym_or_z):
    """Convert element symbol or Z into a valid mixed-case element name.

    Args:
      sym_or_z: string or numeric type representing an element symbol or Z.

    Returns:
      A valid mixed-case element name that matches the symbol or Z.

    Raises:
      ElementNameError: if the symbol or Z cannot be converted.
    """

    try:
        return _NAME_FROM_SYMBOL[validated_symbol(sym_or_z)]
    except ElementSymbolError:
        pass
    try:
        return _NAME_FROM_Z[validated_z(sym_or_z)]
    except ElementZError:
        raise ElementNameError("Must supply either the element symbol or Z")


class Element:
    """Basic properties (symbol, name, Z, and mass) of an element.

    Also provides string formatting:
    >>> elem = Element('Ge')
    >>> '{:%n(%s) Z=%z}'.format(elem)
    'Germanium(Ge) Z=32'

    Properties:
      symbol (read-only): the mixed-case element symbol (e.g., "Ge")
      name (read-only): the mixed-case element name (e.g., "Germanium")
      Z (read-only): an integer giving the atomic number (e.g., 32)
      atomic_mass (read-only): a float giving the atomic mass in amu
    """

    def __init__(self, arg):
        """Instantiate by providing a name, symbol, or Z value.

        Args:
          arg: a numeric or string type giving a name, symbol, or Z value.

        Raises:
          ElementError: if Element could not be instantiated.
        """

        self.symbol = None
        self.name = None
        self.Z = None
        self.atomic_mass = None
        try:
            self._init_sym(arg)
        except ElementSymbolError:
            try:
                self._init_name(arg)
            except ElementNameError:
                try:
                    self._init_z(arg)
                except ElementZError:
                    raise ElementError(f"Could not instantiate Element: {arg}")
        self.atomic_mass = _MASS_FROM_SYMBOL[self.symbol]

    def _init_sym(self, arg):
        """Initialize with an element symbol."""
        self.symbol = validated_symbol(arg)
        self.name = element_name(self.symbol)
        self.Z = element_z(self.symbol)

    def _init_name(self, arg):
        """Initialize with an element name."""
        self.name = validated_name(arg)
        self.symbol = element_symbol(self.name)
        self.Z = element_z(self.name)

    def _init_z(self, arg):
        """Initialize with an element Z."""
        self.Z = validated_z(arg)
        self.symbol = element_symbol(self.Z)
        self.name = element_name(self.Z)

    def __str__(self):
        """Define behavior of str() on this class."""
        return f"{self}"

    def __format__(self, formatstr):
        """Define behavior of string's format method.

        Format codes:
            '%s': element symbol
            '%n': element name
            '%z': element Z
        """

        str0 = str(formatstr)
        if len(str0) == 0:
            str0 = "%n(%s) Z=%z"
        str0 = str0.replace("%s", self.symbol)
        str0 = str0.replace("%n", self.name)
        str0 = str0.replace("%z", f"{self.Z}")
        return str0

    def __eq__(self, other):
        """Define equality of two elements."""
        try:
            return (
                self.name == other.name
                and self.symbol == other.symbol
                and self.Z == other.Z
            )
        except:
            raise ElementError("Cannot determine equality")



"""General utility functions to be shared among core modules."""

import sys
import datetime
from dateutil.parser import parse as dateutil_parse
from uncertainties import UFloat, unumpy
import warnings
import numpy as np

EPS = np.finfo(float).eps

VECTOR_TYPES = (list, tuple, np.ndarray)


class UncertaintiesError(Exception):
    """Raised when uncertainties are badly specified in an input."""

    pass


def all_ufloats(x):
    """Check if every element of x is a UFloat.

    Args:
      x: an iterable or scalar to check

    Raises:
      UncertaintiesError: if some but not all elements of x are UFloats

    Returns:
      True if all elements of x are UFloats.
      False if no elements of x are UFloats.
    """

    try:
        are_ufloats = [isinstance(xi, UFloat) for xi in x]
    except TypeError:
        return isinstance(x, UFloat)
    else:
        if all(are_ufloats):
            return True
        elif any(are_ufloats):
            raise UncertaintiesError("Input should be all UFloats or no UFloats")
        else:
            return False


def handle_uncs(x_array, x_uncs, default_unc_func):
    """Handle two methods of specifying uncertainties (UFloats or manually).

    Args:
      x_array: a list/tuple/array that may contain UFloats
      x_uncs: a list/tuple/array that may contain manual uncertainty values
      default_unc_func: a function that will take as input x_array and
        return a set of default values for x_uncs (for if x_uncs not specified
        and x_array not UFloats)

    Raises:
      UncertaintiesError: if both UFloats and manual uncertainties are
        specified

    Returns:
      a np.array of UFloats
    """

    ufloats = all_ufloats(x_array)

    if ufloats and x_uncs is None:
        return np.asarray(x_array)
    elif ufloats:
        raise UncertaintiesError(
            "Specify uncertainties with UFloats or "
            + "by separate argument, but not both"
        )
    elif x_uncs is not None:
        return unumpy.uarray(x_array, x_uncs)
    else:
        return unumpy.uarray(x_array, default_unc_func(x_array))


def handle_datetime(input_time, error_name="datetime arg", allow_none=False):
    """Parse an argument as a date, datetime, date+time string, or None.

    Args:
      input_time: the input argument to be converted to a datetime
      error_name: the name to be displayed if an error is raised.
        (default: 'datetime arg')
      allow_none: whether a None is allowed as an input and return value.
        (default: False)

    Raises:
      TypeError: if input_time is not a string, datetime, date, or None

    Returns:
      a datetime.datetime, or None
    """

    if isinstance(input_time, datetime.datetime):
        return input_time
    elif isinstance(input_time, datetime.date):
        warnings.warn(
            "datetime.date passed in with no time; defaulting to 0:00 on date"
        )
        return datetime.datetime(input_time.year, input_time.month, input_time.day)
    elif isinstance(input_time, str):
        return dateutil_parse(input_time)
    elif input_time is None and allow_none:
        return None
    else:
        raise TypeError(f"Unknown type for {error_name}: {input_time}")


def bin_centers_from_edges(edges_kev):
    """Calculate bin centers from bin edges.

    Args:
      edges_kev: an iterable representing bin edge values

    Returns:
      np.array of length (len(edges_kev) - 1), representing bin center
        values with the same units as the input
    """

    edges_kev = np.array(edges_kev)
    centers_kev = (edges_kev[:-1] + edges_kev[1:]) / 2
    return centers_kev


def sqrt_bins(bin_edge_min, bin_edge_max, nbins):
    """
    Square root binning

    Args:
      bin_edge_min (float): Minimum bin edge (must be >= 0)
      bin_edge_max (float): Maximum bin edge (must be greater than bin_min)
      nbins (int): Number of bins

    Returns:
      np.array of bin edges (length = nbins + 1)
    """
    assert bin_edge_min >= 0
    assert bin_edge_max > bin_edge_min
    return np.linspace(np.sqrt(bin_edge_min), np.sqrt(bin_edge_max), nbins + 1) ** 2



import numba as nb
import numpy as np
import warnings


class RebinError(Exception):
    """Exception raised by rebin operations."""

    pass


class RebinWarning(UserWarning):
    """Warnings displayed by rebin operations."""

    pass


def _check_monotonic_increasing(arr, arr_name="array"):
    """Check that a numpy array is monotonically increasing.

    Args:
      arr: numpy array for checking
      arr_name: name of array, just for use in the AssertionError message

    Raises:
      AssertionError: if arr is not monotonically increasing
    """
    # Check that elements along the last axis are increasing or
    # neighboring elements are equal
    tmp = np.diff(arr)
    if not ((tmp > 0) | np.isclose(tmp, 0)).all():
        raise RebinError(f"{arr_name} is not monotonically increasing: {arr}")


def _check_partial_overlap(in_edges, out_edges):
    """In and out edges partial overlap checking.

    Args:
        in_edges (np.ndarray): an array of the input bin edges (1D or 2D)
            [num_spectra, num_bins_in + 1] or [num_bins_in + 1]
        out_edges (np.ndarray): an array of the output bin edges
            [num_bins_out]

    Raises:
        RebinWarning: for the following cases:

            old:   └┴┴ ...
            new: └┴┴┴┴ ...

            old: ... ┴┴┴┘
            new: ... ┴┴┴┴┴┘

    """
    if (in_edges[..., 0] > out_edges[..., 0]).any():
        warnings.warn(
            "The first input edge is larger than the first output edge, "
            "zeros will padded on the left side of the new spectrum",
            RebinWarning,
        )
    if (in_edges[..., -1] < out_edges[..., -1]).any():
        warnings.warn(
            "The last input edge is smaller than the last output edge, "
            "zeros will padded on the right side of the new spectrum",
            RebinWarning,
        )


def _check_any_overlap(in_edges, out_edges):
    """In and out edges overlapping at all.

    Args:
        in_edges (np.ndarray): an array of the input bin edges
            [..., num_bins_in + 1] (defined as the last dim in the array)
        out_edges (np.ndarray): an array of the output bin edges
            [num_bins_out]

    Raises:
        RebinError: for the following cases:

            old:   └┴┴┴┴┘
            new:             └┴┴┴┘

            old:           └┴┴┴┴┘
            new:   └┴┴┴┘
    """
    if (in_edges[..., -1] <= out_edges[..., 0]).any():
        raise RebinError("Input edges are all smaller than output edges")
    if (in_edges[..., 0] >= out_edges[..., -1]).any():
        raise RebinError("Input edges are all larger than output edges")


@nb.vectorize([nb.f8(nb.f8, nb.f8, nb.f8, nb.f8)], nopython=True)
def _linear_offset(slope, cts, low, high):
    """
    Calculate the offset of the linear approximation of slope when splitting
    counts between bins.

    Args:
      slope:
      cts: counts within the bin
      low: lower bin edge energy
      high: higher bin edge energy

    Returns:
      the offset
    """
    if np.abs(slope) < 1e-6:
        offset = cts / (high - low)
    else:
        offset = (cts - slope / 2.0 * (high ** 2 - low ** 2)) / (high - low)
    return offset


@nb.vectorize([nb.f8(nb.f8, nb.f8, nb.f8)], nopython=True)
def _slope_integral(x, m, b):
    """Indefinite integral of y = mx + b, with an x value substituted in.

    Args:
      x: the x-value
      m: the value of the slope
      b: the y-offset value

    Returns:
      The indefinite integral of y = mx + b, with an x value substituted in.
      (m x^2 / 2 + b x)
    """
    return m * x ** 2 / 2 + b * x


@nb.vectorize([nb.f8(nb.f8, nb.f8, nb.f8, nb.f8)], nopython=True)
def _counts(m, b, x_low, x_high):
    """Definite integral of y = mx + b.

    Computes the area under a linear approximation of the changing count
    rate in the vincity of relevant bins.  Edges of this integration
    are low and high while offset is provided from _linear_offset and cts
    from the bin being partitioned.

    Args:
      m: the value of slope m
      b: the value of y-offset b
      x_low: the "high" value for x, to be substituted into the integral
      x_high: the "low" value for x, to be substituted into the integral

    Returns:
      the definite integral of y = mx + b
    """
    return _slope_integral(x_high, m, b) - _slope_integral(x_low, m, b)


@nb.guvectorize(
    [(nb.f8[:], nb.f8[:], nb.f8[:], nb.f8[:], nb.f8[:])], "(n),(N),(m),(n)->(m)"
)
def _rebin_interpolation(
    in_spectrum, in_edges, out_edges_no_rightmost, slopes, out_spectrum
):
    """Rebins a spectrum using linear interpolation.

    Keeps a running counter of two loop indices: in_idx & out_idx

    N.B. Does not keep Poisson statistics

    Args:
      in_spectrum: iterable of input spectrum counts in each bin
      in_edges: iterable of input bin edges (len = len(in_spectrum) + 1)
      out_edges_no_rightmost: iterable of output bin edges
                              (sans the rightmost bin)
      slopes: the slopes of each histogram bin, with the lines drawn between
              each bin edge. (len = len(in_spectrum))
      out_spectrum: for nb.guvectorize; This is the return array, do not
                    actually give this as an arg when calling the function.
                    Same length as out_edges_no_rightmost due to signature.

    Returns:
      1D numpy array of rebinned spectrum counts in each bin
    """
    # N.B. use [:] to assign values to full array output with guvectorize:
    out_spectrum[:] = np.zeros_like(out_spectrum)  # init output
    # out_edges: might not actually need to create this, since for loops
    # handles under/over-flows.
    # TODO check logic and do whichever is the least computationally expensive
    out_edges = np.concatenate((out_edges_no_rightmost, np.array([np.inf])))
    # in_idx: input bin or left edge
    #         init to the first in_bin which overlaps the 0th out_bin
    in_idx = max(0, np.searchsorted(in_edges, out_edges[0]) - 1)
    # Under-flow handling: Put all counts from in_bins that are completely to
    # the left of the 0th out_bin into the 0th out_bin
    out_spectrum[0] += np.sum(in_spectrum[:in_idx])
    # out_idx: output bin or left edge
    #          init to the first out_bin which overlaps the 0th in_bin
    out_idx = max(0, np.searchsorted(out_edges, in_edges[0]) - 1)
    # loop through input bins (starting from the above in_idx value):
    for in_idx in range(in_idx, len(in_spectrum)):
        # input bin info/data
        in_left_edge = in_edges[in_idx]
        in_right_edge = in_edges[in_idx + 1]
        counts = in_spectrum[in_idx]
        slope = slopes[in_idx]
        offset = _linear_offset(slope, counts, in_left_edge, in_right_edge)
        # loop through output bins that overlap with the current input bin
        for out_idx in range(out_idx, len(out_spectrum)):
            out_left_edge = out_edges[out_idx]
            out_right_edge = out_edges[out_idx + 1]
            if out_left_edge > in_right_edge:
                out_idx -= 1  # rewind back to previous out_bin; not done yet
                break  # break out of out_bins loop, move on to next in_bin
            # Low edge for interpolation
            if out_idx == 0:
                # under-flow handling:
                # all counts in this in_bin goes into the 0th out_bin
                low = in_left_edge  # == min(in_left_edge, out_edges[0])
            else:
                low = max(in_left_edge, out_left_edge)
            # High edge for interpolation
            if out_idx == len(out_spectrum) - 1:
                # over-flow handling:
                # all counts in this in_bin goes into this out_bin
                high = in_right_edge
            else:
                high = min(in_right_edge, out_right_edge)
            # Calc counts for this bin
            out_spectrum[out_idx] += _counts(slope, offset, low, high)


@nb.guvectorize([(nb.i8[:], nb.f8[:], nb.f8[:], nb.i8[:])], "(n),(N),(m)->(m)")
def _rebin_listmode(in_spectrum, in_edges, out_edges_no_rightmost, out_spectrum):
    """Stochastic rebinning method: spectrum-histogram to listmode then back.

    rightmost edge in the parameter out_edges chopped off, in order for
    nb.guvectorize to work, because it requires the output dimensions
    to be specified by the dimensions of the input parameters
    (i.e. m -> m-1 is not allowed).
    This works since we put all overflow values in the leftmost and rightmost
    bins anyways.

    TODO Assume piecewise constant (ie steps, flat within each bin)
         distribution for in_spectrum (for now). slopes is unused.

    Args:
      in_spectrum: iterable of input spectrum counts in each bin
      in_edges: iterable of input bin edges (len = len(in_spectrum) + 1)
      out_edges_no_rightmost: iterable of output bin edges
                              (sans the rightmost bin)
      out_spectrum: for nb.guvectorize; This is the return array, do not
                    actually give this as an arg when calling the function.
                    Same length as out_edges_no_rightmost due to signature.

    Returns:
      1D numpy array of rebinned spectrum counts in each bin
    """
    # knock out leftmost bin edge too, because we put all overflows into
    # first and last bins anyways
    out_edges = np.concatenate(
        (np.array([-np.inf]), out_edges_no_rightmost[1:], np.array([np.inf]))
    )
    energies = np.zeros(np.sum(in_spectrum))
    energy_idx_start = 0
    # loop through input bins:
    # in_idx = index of in_bin or in_left_edge
    for in_idx in range(len(in_spectrum)):
        bin_counts = in_spectrum[in_idx]
        energy_idx_stop = energy_idx_start + bin_counts
        in_left_edge = in_edges[in_idx]
        in_right_edge = in_edges[in_idx + 1]
        energies[energy_idx_start:energy_idx_stop] = in_left_edge + (
            in_right_edge - in_left_edge
        ) * np.random.rand(bin_counts)
        energy_idx_start = energy_idx_stop
    # bin the energies (drops the energies outside of the out-binning-range)
    # N.B. use [:] to assign values to full array output with guvectorize:
    out_spectrum[:] = np.histogram(energies, bins=out_edges)[0]


def rebin(
    in_spectra,
    in_edges,
    out_edges,
    method="interpolation",
    slopes=None,
    zero_pad_warnings=True,
):
    """
    Spectra rebinning via deterministic or stochastic methods.

    Args:
        in_spectrum (np.ndarray): an ND array of input counts spectra
            [..., num_bins_in]
        in_edges (np.ndarray): an array of the input bin edges
            [..., num_bins_in + 1] or [num_bins_in + 1] (either contains
            different bin edges for each element in in_spectrum or a single
            array that applies to all elements in in_spectrum)
        out_edges (np.ndarray): an array of the output bin edges
            [..., num_bins_out + 1] or [num_bins_out + 1] (either contains
            different bin edges for each element in in_spectrum or a single
            array that applies to all elements in in_spectrum)
        method (str): rebinning method
            "interpolation"
                Deterministic interpolation
            "listmode"
                Stochastic rebinning via conversion to listmode of energies.
                This method will internally convert input spectrum values to
                integers (if necessary) and raise a RebinWarning if the
                conversion results in a decimal precision loss.
        slopes (np.ndarray|None): (optional) an array of input bin slopes for
            quadratic interpolation (1D or 2D)
            (only applies for "interpolation" method)
            [num_spectra, num_bins_in + 1] or [num_bins_in + 1]
        zero_pad_warnings (boolean): warn when edge overlap results in
            appending empty bins

    Raises:
        AssertionError: for bad input arguments

    Returns:
        The rebinned spectrum/a
    """
    method = method.lower()

    # Convert inputs to numpy types
    in_edges = np.asarray(in_edges)
    out_edges = np.asarray(out_edges)
    in_spectra = np.asarray(in_spectra)
    # Cast data types and check listmode input
    if method == "listmode":
        if (in_spectra < 0).any():
            raise RebinError(
                "Cannot rebin spectra with negative values with listmode method"
            )
        if (in_spectra < 1).all():
            raise RebinError(
                "Cannot rebin spectra with all values less than "
                "one with listmode method"
            )
        if np.issubdtype(in_spectra.dtype.type, np.floating):
            # np.rint is a ufunc: allows using this with tools such as xarray
            in_spectra_rint = np.rint(in_spectra).astype(int)
            if np.allclose(in_spectra, in_spectra_rint):
                # Don't warn in the case of floats which round to integers
                pass
            else:
                warnings.warn(
                    "Argument in_spectra contains float value(s) which "
                    + "will have decimal precision loss when converting to "
                    + "integers for rebin method listmode.",
                    RebinWarning,
                )
            in_spectra = in_spectra_rint

    if slopes is not None:
        slopes = np.asarray(slopes)
        # if slope is wrong dimension error will be raised in guvectorize
    elif method == "interpolation":  # "listmode" doesn't use slopes for now
        slopes = np.zeros(in_spectra.shape[-1])

    # Only check that in_spectra and in_edges are compatible any other shape
    # issue should raise an error in guvectorize
    assert in_spectra.shape[-1] == in_edges.shape[-1] - 1
    # Check for increasing bin structure
    _check_monotonic_increasing(in_edges, "in_edges")
    _check_monotonic_increasing(out_edges, "out_edges")
    # Check for bin structure overlap
    _check_any_overlap(in_edges, out_edges)
    if zero_pad_warnings:
        _check_partial_overlap(in_edges, out_edges)
    # Remove highest output edge, necessary for guvectorize
    out_edges = out_edges[..., :-1]

    # Specific calls to (wrapped) nb.guvectorize'd rebinning methods
    if method == "interpolation":
        return _rebin_interpolation(in_spectra, in_edges, out_edges, slopes)
    elif method == "listmode":
        return _rebin_listmode(in_spectra, in_edges, out_edges)
    raise ValueError(f"{method} is not a valid rebinning method")



""""Energy calibration classes"""

from abc import ABCMeta, abstractmethod, abstractproperty
from future.builtins import dict, super, zip
from future.utils import viewitems
import numpy as np
import matplotlib.pyplot as plt
import warnings


class EnergyCalError(Exception):
    """Base class for errors in energycal.py"""

    pass


class BadInput(EnergyCalError):
    """Error related to energy cal input"""

    pass


class EnergyCalBase:
    """Abstract base class for energy calibration.

    A note on nomenclature: for historic reasons, 'channels' is used in
    energycal.py for generic uncalibrated x-axis values. A 'channel' is no
    longer necessarily an integer channel number (i.e., bin) from a
    multi-channel analyzer, but could for instance be a float-type fC of charge
    collected.

    Subclasses must implement:
      _ch2kev (method)
      kev2ch (method)
      valid_coeffs (property)
      _perform_fit (method)
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        """Create an empty calibration instance.

        Normally you should use from_points or from_coeffs classmethods.

        Args:
          none
        """

        self._calpoints = dict()
        self._coeffs = dict()
        # initialize fit constraints?
        warnings.warn(
            "The use of bq.EnergyCalBase classes is deprecated "
            "and will be removed in a future release; "
            "use bq.Calibration instead",
            DeprecationWarning,
        )

    @classmethod
    def from_points(cls, chlist, kevlist, include_origin=False):
        """Construct EnergyCal from calibration points.

        Args:
          chlist: list/tuple/array of the channel values of calibration points
          kevlist: list/tuple/array of the corresponding energy values [keV]
          include_origin: Default is False, if set to True will add a point
                          at zero into the fit.

        Raises:
          BadInput: for bad chlist and/or kevlist.
        """

        if chlist is None or kevlist is None:
            raise BadInput("Channel list and energy list are required")

        try:
            cond = len(chlist) != len(kevlist)
        except TypeError:
            raise BadInput("Inputs must be one dimensional iterables")
        if cond:
            raise BadInput("Channels and energies must be same length")

        cal = cls()

        if include_origin:
            cal.new_calpoint(0, 0)

        for ch, kev in zip(chlist, kevlist):
            try:
                cal.new_calpoint(ch, kev)
            except (ValueError, TypeError):
                raise BadInput("Inputs must be one dimensional iterables")
        cal.update_fit()
        return cal

    @classmethod
    def from_coeffs(cls, coeffs):
        """Construct EnergyCal from equation coefficients dict.

        Args:
          coeffs: a dict with keys equal to elements in valid_coeffs,
            and values specifying the value of the coefficient
        """

        cal = cls()

        for coeff, val in viewitems(coeffs):
            cal._set_coeff(coeff, val)

        # TODO make sure all coefficients are specified

        return cal

    @property
    def channels(self):
        """The channel values of calibration points.

        Returns:
          an np.ndarray of channel values (floats)
        """

        return np.array(list(self._calpoints.values()), dtype=float)

    @property
    def energies(self):
        """The energy values of calibration points.

        Returns:
          an np.ndarray of energy values [keV]
        """

        return np.array(list(self._calpoints), dtype=float)

    @property
    def calpoints(self):
        """The calibration points, in (ch, kev) pairs.

        Returns:
          a list of 2-element tuples of (channel, energy[keV])
        """

        return list(zip(self.channels, self.energies))

    @property
    def coeffs(self):
        """The coefficients of the current calibration curve.

        Returns:
          a dict of {coeff: value}
        """

        # TODO: if there are no coeffs, error?
        return self._coeffs

    def add_calpoint(self, ch, kev):
        """Add a calibration point (ch, kev) pair. May be new or existing.

        Args:
          ch: the channel value of the calibration point
          kev: the energy value of the calibration point [keV]
        """

        self._calpoints[float(kev)] = float(ch)

    def new_calpoint(self, ch, kev):
        """Add a new calibration point. Error if energy matches existing point.

        Args:
          ch: the channel value of the calibration point
          kev: the energy value of the calibration point [keV]

        Raises:
          EnergyCalError: if energy value already exists in calibration
        """

        if kev in self._calpoints:
            raise EnergyCalError("Calibration energy already exists")
        self.add_calpoint(ch, kev)

    def rm_calpoint(self, kev):
        """Remove a calibration point, if it exists.

        Args:
          the energy value of the point to remove [keV]
        """

        if kev in self._calpoints:
            del self._calpoints[kev]
        # TODO erroring version?

    def ch2kev(self, ch):
        """Convert channel(s) to energy value(s).

        Args:
          ch: a scalar, np.array, list or tuple of channel values

        Returns:
          the energy value(s) corresponding to the channel value(s) [keV].
            a float if input is scalar. an np.array if input is iterable
        """

        if isinstance(ch, (list, tuple)):
            ch = np.array(ch)

        return self._ch2kev(ch)

    @abstractmethod
    def _ch2kev(self, ch):
        """Convert scalar OR np.array of channel(s) to energies.

        Should use numpy ufuncs so that the input dtype doesn't matter.

        Args:
          ch: an np.array, float, or int of channel values

        Returns:
          energy values, the same size/type as ch [keV]
        """

        pass

    def kev2ch(self, kev):
        """Convert energy value(s) to channel(s).

        Args:
          kev: a scalar, np.array, list or tuple of energy values [keV]

        Returns:
          the channel value(s) corresponding to the input energies.
            a float if input is scalar. an np.array if input is iterable
        """

        if isinstance(kev, (list, tuple)):
            kev = np.array(kev)

        return self._kev2ch(kev)

    @abstractmethod
    def _kev2ch(self, kev):
        """Convert energy value(s) to channel(s).

        Should use numpy ufuncs so that the input dtype doesn't matter.

        Args:
          kev: an np.array, float, or int of energy values [keV]

        Returns:
          the channel value(s) corresponding to the input energies.
            a float if input is scalar. an np.array if input is iterable
        """

        # if this is not possible, raise a NotImplementedError ?
        pass

    @abstractproperty
    def valid_coeffs(self):
        """A list of valid coefficients for the calibration curve.

        Returns:
          a tuple of strings, the names of the coefficients for this curve
        """

        pass

    def _set_coeff(self, name, val):
        """Set a coefficient for the calibration curve.

        Args:
          name: a string, the name of the coefficient to set
          val: the value to set the coefficient to

        Raises:
          EnergyCalError: if name is not in valid_coeffs
        """

        if name in self.valid_coeffs:
            self._coeffs[name] = val
        else:
            raise EnergyCalError(f"Invalid coefficient name: {name}")

    def update_fit(self):
        """Compute the calibration curve from the current points.

        Raises:
          EnergyCalError: if there are too few calibration points to fit
        """

        num_coeffs = len(self.valid_coeffs)
        # TODO: free coefficients, not all coefficients
        num_points = len(self._calpoints)

        if num_points == 0:
            raise EnergyCalError("No calibration points; cannot calibrate")
        elif num_points < num_coeffs:
            raise EnergyCalError("Not enough calibration points to fit curve")
        else:
            self._perform_fit()

    @abstractmethod
    def _perform_fit(self):
        """Do the actual curve fitting."""

        pass

    def plot(self, ax=None):
        """Plot the calibration.

        Parameters
        ----------
        ax : np.ndarray of shape (2,), or matplotlib axes object, optional
            Plot axes to use. If None, create new axes.
        """

        # Handle whether we have fit points or just a fit function
        has_points = self.channels.size > 0

        if ax is None:
            fig, ax = plt.subplots(1 + has_points, 1, sharex=True)

        if has_points:
            assert ax.shape == (2,)
            ax_cal, ax_res = ax
            xmin, xmax = self.channels.min(), self.channels.max()
        else:
            ax_cal = ax
            xmin, xmax = 0, 3000

        # Plot calibration curve
        xx = np.linspace(xmin, xmax, 1000)
        yy = self.ch2kev(xx)
        ax_cal.plot(xx, yy, alpha=1.0 - 0.7 * has_points)
        ax_cal.set_ylabel("energy [keV]")

        if has_points:
            # Plot calibration points
            ax_cal.scatter(self.channels, self.energies)

            # Plot residuals
            ax_res.scatter(self.channels, self.ch2kev(self.channels) - self.energies)
            ax_res.set_xlabel("channel")
            ax_res.set_ylabel("fit-data [keV]")
            ax_res.axhline(0, linestyle="dashed", linewidth=1, c="k")
        else:
            ax_cal.set_xlabel("channel")


# TODO: dummy class for testing?


class LinearEnergyCal(EnergyCalBase):
    """
    kev = b*ch + c
    """

    @classmethod
    def from_coeffs(cls, coeffs):
        """Construct LinearEnergyCal from equation coefficients dict.

        Valid coefficient names (slope, offset):
          ('b', 'c')
          ('p1', 'p0')
          ('slope', 'offset')
          ('m', 'b')

        Args:
          coeffs: a dict with keys equal to valid coeff names,
            and values specifying the value of the coefficient
        """

        new_coeffs = {}
        if "p0" in coeffs and "p1" in coeffs:
            new_coeffs["b"] = coeffs["p1"]
            new_coeffs["c"] = coeffs["p0"]
        elif "slope" in coeffs and "offset" in coeffs:
            new_coeffs["b"] = coeffs["slope"]
            new_coeffs["c"] = coeffs["offset"]
        elif "m" in coeffs and "b" in coeffs:
            new_coeffs["b"] = coeffs["m"]
            new_coeffs["c"] = coeffs["b"]
        else:
            new_coeffs = coeffs.copy()
        cal = super().from_coeffs(new_coeffs)
        return cal

    @property
    def valid_coeffs(self):
        """A list of valid coefficients for the calibration curve.

        Returns:
          a tuple of strings, the names of the coefficients for this curve
        """

        return ("b", "c")

    @property
    def slope(self):
        """Return the slope coefficient value."""

        try:
            return self._coeffs["b"]
        except KeyError:
            raise EnergyCalError("Slope coefficient not yet supplied or calculated.")

    @property
    def offset(self):
        """Return the offset coefficient value."""

        try:
            return self._coeffs["c"]
        except KeyError:
            raise EnergyCalError("Offset coefficient not yet supplied or calculated.")

    def _ch2kev(self, ch):
        """Convert scalar OR np.array of channel(s) to energies.

        Should use numpy ufuncs so that the input dtype doesn't matter.

        Args:
          ch: an np.array, float, or int of channel values

        Returns:
          energy values, the same size/type as ch [keV]
        """

        return self.slope * ch + self.offset

    def _kev2ch(self, kev):
        """Convert energy value(s) to channel(s).

        Args:
          kev: an np.array, float, or int of energy values [keV]

        Returns:
          the channel value(s) corresponding to the input energies.
            a float if input is scalar. an np.array if input is iterable
        """

        return (kev - self.offset) / self.slope

    def _perform_fit(self):
        """Do the actual curve fitting."""

        b, c = np.polyfit(self.channels, self.energies, 1)
        self._set_coeff("b", b)
        self._set_coeff("c", c)



"""Class to describe a generic calibration function."""

from abc import abstractmethod
import ast
import copy
import asteval
import black
import blib2to3
import numpy as np
import scipy
import scipy.optimize
import matplotlib.pyplot as plt
from .. import io

DEFAULT_DOMAIN = (0, 1e5)
DEFAULT_RANGE = (0, 1e5)

safe_eval = asteval.Interpreter(use_numpy=False)
safe_eval.symtable["np"] = np
safe_eval.symtable["numpy"] = np
safe_eval.symtable["scipy"] = scipy


class CalibrationError(Exception):
    """Base class for calibration errors."""

    pass


def _validate_domain_range(domain, rng):
    """Validate that domain and range contain finite values.

    Parameters
    ----------
    domain : array_like
        The domain of the function. Will raise an error if the independent
        variable is outside this interval. Must be finite.
        By default DEFAULT_DOMAIN.
    rng : array_like
        The range of the function. Expression outputs will be clipped to this
        interval. Must be finite. By default DEFAULT_RANGE.
    """
    # must be length-2 iterables
    try:
        len(domain)
    except TypeError:
        raise CalibrationError(f"Domain must be length-2 iterable: {domain}")
    domain = np.asarray(domain)
    if not (len(domain) == 2 and domain.ndim == 1):
        raise CalibrationError(f"Domain must be length-2 iterable: {domain}")
    try:
        len(rng)
    except TypeError:
        raise CalibrationError(f"Range must be length-2 iterable: {rng}")
    rng = np.asarray(rng)
    if not (len(rng) == 2 and rng.ndim == 1):
        raise CalibrationError(f"Range must contain two values: {rng}")
    # must contain finite values
    if not np.isfinite(domain[0]) or not np.isfinite(domain[1]):
        raise CalibrationError(f"Domain must contain finite values: {domain}")
    if not np.isfinite(rng[0]) or not np.isfinite(rng[1]):
        raise CalibrationError(f"Range must contain finite values: {rng}")
    # must be in ascending order
    if not (domain[1] > domain[0]):
        raise CalibrationError(f"Domain must contain ascending values: {domain}")
    if not (rng[1] > rng[0]):
        raise CalibrationError(f"Range must contain ascending values: {rng}")


def _eval_expression(
    expression,
    params,
    x,
    ind_var="x",
    aux_params=None,
    domain=DEFAULT_DOMAIN,
    rng=DEFAULT_RANGE,
):
    """Evaluate the expression at x.

    Parameters
    ----------
    expression : string
        The expression that defines the calibration function.
    params : array_like
        List of floating point parameters for the calibration function,
        referred to by the symbol "p".
    x : float or array_like
        The argument at which to evaluate the expression.
    ind_var : str
        The symbol of the independent variable. Default "x", "y" also allowed.
    aux_params : array_like
        Auxiliary floating point parameters for the calibration function,
        referred to by the symbol "a". By default an empty array.
    domain : array_like
        The domain of the function. Will raise an error if the independent
        variable is outside this interval. Must be finite. By default
        DEFAULT_DOMAIN.
    rng : array_like
        The range of the function. Expression outputs will be clipped to this
        interval. Must be finite. By default DEFAULT_RANGE.

    Returns
    -------
    y : float or array_like
        Result of evaluating the expression for x.
    """
    _validate_domain_range(domain, rng)
    x = np.asarray(x)
    if not np.all(x >= domain[0]):
        raise CalibrationError(f"{ind_var} must be >= {domain[0]}: {x}")
    if not np.all(x <= domain[1]):
        raise CalibrationError(f"{ind_var} must be <= {domain[1]}: {x}")
    if ind_var not in ["x", "y"]:
        raise CalibrationError(f"Independent variable {ind_var} must be 'x' or 'y'")
    if aux_params is None:
        aux_params = np.array([])
    else:
        aux_params = np.asarray(aux_params)
    safe_eval.symtable["p"] = params
    safe_eval.symtable["a"] = aux_params
    safe_eval.symtable[ind_var] = x
    y = safe_eval(expression)
    if len(safe_eval.error) > 0:
        raise CalibrationError(
            "asteval failed with errors:\n"
            + "\n".join(str(err.get_error()) for err in safe_eval.error)
        )
    if not np.all(np.isreal(y)):
        raise CalibrationError(f"Function evaluation resulted in complex values: {y}")
    # clip values of y to the range
    y = np.clip(y, rng[0], rng[1])
    return y


def _param_indices(expression):
    """Find all integer parameter indices of the expression.

    The expression must explicitly call each parameter as "p[j]", where
    j is the index of the parameter.

    Parameters
    ----------
    expression : string
        The expression that defines the calibration function.

    Returns
    -------
    param_indices : array_like
        List of integer parameter indices appearing in the expression.
    """
    # find parameter indices
    tokens = expression.split("p[")
    param_indices = [int(token.split("]")[0]) for token in tokens[1:]]
    param_indices = np.array(sorted(np.unique(param_indices)))
    return param_indices


def _validate_expression(
    expression,
    ind_var="x",
    params=None,
    aux_params=None,
    domain=DEFAULT_DOMAIN,
    rng=DEFAULT_RANGE,
    n_eval=100,
):
    """Perform checks on the expression.

    The expression must explicitly call each parameter as "p[j]", where
    j is the index of the parameter, and the indices for n parameters
    range from 0 to n - 1. The expression is checked for how many
    parameters there are and their length is checked if `params` is given
    to ensure each is used at least once.

    Parameters
    ----------
    expression : string
        The expression that defines the calibration function. It will
        be checked for syntax, whether it uses all the parameters,
        and whether it can be evaluated.
    ind_var : str
        The symbol of the independent variable. Default "x", "y" also allowed.
    params : array_like
        List of floating point parameters for the calibration function,
        referred to by the symbol "p". The expression will be checked whether
        it includes all of the parameters.
    aux_params : array_like
        Auxiliary floating point parameters for the calibration function,
        referred to by the symbol "a". By default an empty array.
    domain : array_like
        The domain of the function. Will draw test values from inside this
        interval. Must be finite. By default DEFAULT_DOMAIN.
    rng : array_like
        The range of the function. Expression outputs will be clipped to this
        interval. Must be finite. By default DEFAULT_RANGE.
    n_eval : int
        Number of points in the domain the evaluate at for testing.

    Returns
    -------
    expression : string
        Expression having been validated and reformatted using black.
    """
    _validate_domain_range(domain, rng)

    # apply black formatting for consistency and error checking
    try:
        expression = black.format_str(expression, mode=black.FileMode())
    except (black.InvalidInput, blib2to3.pgen2.tokenize.TokenError):
        raise CalibrationError(
            f"Error while running black on expression:\n{expression}"
        )

    # make sure `ind_var` appears in the formula
    if ind_var not in ["x", "y"]:
        raise CalibrationError(f"Independent variable {ind_var} must be 'x' or 'y'")
    ind_var_appears = False
    for node in ast.walk(ast.parse(expression)):
        if type(node) is ast.Name:
            if node.id == ind_var:
                ind_var_appears = True
    if not ind_var_appears:
        raise CalibrationError(
            f'Independent variable "{ind_var}" must appear in the expression:\n{expression}'
        )

    # make sure each parameter appears at least once
    try:
        param_indices = _param_indices(expression)
    except ValueError:
        raise CalibrationError(
            f"Unable to extract indices to parameters:\n{expression}"
        )
    if len(param_indices) > 0:
        if param_indices.min() != 0:
            raise CalibrationError(
                f"Minimum parameter index in expression is not 0:\n{expression}\n{param_indices}"
            )
        if not np.allclose(np.diff(param_indices), 1):
            raise CalibrationError(
                f"Parameter indices in expression are not contiguous:\n{expression}\n{param_indices}"
            )
    if params is not None:
        if len(param_indices) != len(params):
            raise CalibrationError(
                f"Not enough parameter indices in expression:\n{expression}\n{param_indices}"
            )

    # make sure the expression can be evaluated
    if params is not None:
        x_arr = np.linspace(domain[0], domain[1], num=n_eval)
        for x_val in x_arr:
            try:
                y = _eval_expression(
                    expression,
                    params,
                    x_val,
                    ind_var=ind_var,
                    aux_params=aux_params,
                    domain=domain,
                    rng=rng,
                )
            except CalibrationError:
                raise CalibrationError(
                    f"Cannot evaluate expression for float {ind_var} = {x_val}:\n{expression}\n{safe_eval.symtable['x']}"
                )
        try:
            y = _eval_expression(
                expression,
                params,
                x_arr,
                ind_var=ind_var,
                aux_params=aux_params,
                domain=domain,
                rng=rng,
            )
        except CalibrationError:
            raise CalibrationError(
                f"Cannot evaluate expression for array {ind_var} = {x_arr}:\n{expression}\n{safe_eval.symtable['x']}"
            )

    return expression.strip()


def _fit_expression(
    expression,
    points_x,
    points_y,
    weights=None,
    params0=None,
    aux_params=None,
    domain=DEFAULT_DOMAIN,
    rng=DEFAULT_RANGE,
    **kwargs,
):
    """Fit the expression using the calibration points.

    Performs least squares via scipy.optimize.least_squares.

    Parameters
    ----------
    expression : string
        The expression that defines the calibration function.
    points_x : float or array_like
        The x-value or values of calibration points
    points_y : float or array_like
        The y-value or values of calibration points
    weights : float or array_like
        The weights to be used when fitting. Fitting will minimize the sum
        of squared errors weighted by these weights, i.e.,
        sum_i (w_i*(y_i - f(x_i))**2)). By default will use an array of
        ones with the same length as the points.
    params0 : float or array_like
        Initial guesses for the parameters. By default an array of ones
        with its length inferred from the number of parameters
        referenced in the expression.
    aux_params : array_like
        Auxiliary floating point parameters for the calibration function,
        referred to by the symbol "a". By default an empty array.
    domain : array_like
        The domain of the function. Will raise an error if the independent
        variable is outside this interval. Must be finite. By default
        DEFAULT_DOMAIN.
    rng : array_like
        The range of the function. Expression outputs will be clipped to this
        interval. Must be finite. By default DEFAULT_RANGE.
    kwargs : dict
        Kwargs to pass to the minimization routine.

    Returns
    -------
    params : array_like
        Parameters that result from the fit.
    """
    expression = _validate_expression(
        expression, aux_params=aux_params, domain=domain, rng=rng
    )
    points_x, points_y, weights = _check_points(
        points_x, points_y, weights=weights, domain=domain, rng=rng
    )

    # check that we have the expected number of parameters
    n_params = len(_param_indices(expression))
    if params0 is None:
        params0 = np.ones(n_params)
    else:
        params0 = np.asarray(params0).flatten()
    if len(params0) != n_params:
        raise CalibrationError(
            f"Starting parameters have length {len(params0)}, but expression requires {n_params} parameters"
        )
    expression = _validate_expression(
        expression, params=params0, aux_params=aux_params, domain=domain, rng=rng
    )

    # check that we have enough points
    if len(points_x) < n_params:
        raise CalibrationError(
            f"Expression has {n_params} free parameters but there are only {len(points_x)} points to fit"
        )

    # skip fitting if there are zero parameters to fit
    if n_params == 0:
        return np.array([])

    # define the residuals for least squares
    def residuals(p, xs, ys, ws):
        fs = _eval_expression(
            expression, p, xs, aux_params=aux_params, domain=domain, rng=rng
        )
        return np.sqrt(ws) * (ys - fs)

    # perform the fit
    results = scipy.optimize.least_squares(
        residuals,
        params0,
        args=(points_x, points_y, weights),
        **kwargs,
    )
    if not results.success:
        raise CalibrationError(results.message)
    params = results.x
    return params


def _check_points(
    points_x, points_y, weights=None, domain=DEFAULT_DOMAIN, rng=DEFAULT_RANGE
):
    """Perform various checks on the sets of calibration points.

    Ensure the arrays of points are both 1-D and have the same length,
    that all values are >= 0, that they fall within the domain and range,
    and then put them in the order of ascending x values.

    Parameters
    ----------
    points_x : float or array_like
        The x-value or values of calibration points
    points_y : float or array_like
        The y-value or values of calibration points
    weights : float or array_like
        The weights to be used when fitting. Fitting will minimize the sum
        of squared errors weighted by these weights, i.e.,
        sum_i (w_i*(y_i - f(x_i))**2)). By default will use an array of
        ones with the same length as the points.
    domain : array_like
        The domain of the function. Will raise an error if the independent
        variable is outside this interval. Must be finite. By default
        DEFAULT_DOMAIN.
    rng : array_like
        The range of the function. Expression outputs will be clipped to this
        interval. Must be finite. By default DEFAULT_RANGE.

    Returns
    -------
    points_x : array_like
        The x-value or values of calibration points
    points_y : array_like
        The y-value or values of calibration points
    weights : array_like
        The weights to be used when fitting to the points.
    """
    if points_x is None:
        points_x = []
    if points_y is None:
        points_y = []
    points_x = np.atleast_1d(points_x)
    points_y = np.atleast_1d(points_y)
    if points_x.ndim != 1:
        raise CalibrationError(f"Calibration x points must be 1-D: {points_x}")
    if points_y.ndim != 1:
        raise CalibrationError(f"Calibration y points must be 1-D: {points_y}")
    if len(points_x) != len(points_y):
        raise CalibrationError(
            f"Number of x and y calibration points must match: {len(points_x)}, {len(points_y)}"
        )
    # handle the weights
    if weights is None:
        weights = np.ones_like(points_x)
    else:
        weights = np.atleast_1d(weights)
    if weights.ndim != 1:
        raise CalibrationError(f"Calibration weights must be 1-D: {weights}")
    if len(weights) != len(points_x):
        raise CalibrationError(
            f"Number of weights must match points: {len(weights)}, {len(points_x)}"
        )
    if (weights < 0.0).any():
        raise CalibrationError(f"Weights cannot be negative: {weights}")
    # sort points in increasing order of x
    i = np.argsort(points_x)
    points_x = points_x[i]
    points_y = points_y[i]
    weights = weights[i]
    # check domain and range
    if np.any((points_x < domain[0]) | (domain[1] < points_x)):
        raise CalibrationError(
            f"Some x points are outside of domain {domain}: {points_x}"
        )
    if np.any((points_y < rng[0]) | (rng[1] < points_y)):
        raise CalibrationError(f"Some y points are outside of range {rng}: {points_y}")
    return points_x, points_y, weights


def _polynomial_expression(params):
    """Create a polynomial expression of any order.

    The calibration function expression is
        "p[0] + p[1] * x + p[2] * x**2 + ..."

    Parameters
    ----------
    params : array_like
        Coefficients beginning with 0th order.

    Returns
    -------
    expression : str
        The polynomial expression.
    """
    order = len(params) - 1
    if order <= 0:
        raise CalibrationError("Polynomial expression expects an order of at least 1")
    expr = "p[0]"
    for n in range(1, order + 1):
        expr += f" + p[{n}] * x ** {n}"
    return expr


class Calibration:
    """Base class for calibrations.

    A calibration is a scalar function of a scalar argument, parametrized by
    an array of scalars. Examples of calibrations are energy calibrations
    (mapping raw channels to energy in keV), energy resolution calibrations
    (mapping energy to energy FWHM or sigma), and efficiency calibrations
    (mapping energy to fraction of photopeak counts detected).
    """

    def __init__(
        self,
        expression: str,
        params,
        aux_params=None,
        inv_expression=None,
        domain=DEFAULT_DOMAIN,
        rng=DEFAULT_RANGE,
        **attrs,
    ):
        """Create a calibration described by the expression and parameters.

        Parameters
        ----------
        expression : string
            The expression that defines the calibration function as a
            function of argument "x". Parameters are referenced as "p",
            i.e., "p[j]" is the jth parameter, and all parameters must be
            explicitly indexed in the expression. Can be a single-line formula
            like "p[0] + p[1] * x" or a code block.
        params : array_like
            List of floating point parameters for the calibration function
        aux_params : array_like
            Auxiliary floating point parameters for the calibration function,
            referred to by the symbol "a". By default an empty array.
        inv_expression : str
            String giving the inverse of the function. If such an expression
            is available in closed form, it will speed up calls to inverse().
        domain : array_like
            The domain of the function. Will raise an error if the independent
            variable is outside this interval. Must be finite. By default
            DEFAULT_DOMAIN.
        rng : array_like
            The range of the function. Expression outputs will be clipped to
            this interval. Must be finite. By default DEFAULT_RANGE.
        attrs : dict
            Other information to be stored with the calibration.
        """
        self.params = params
        self.aux_params = aux_params
        self.domain = domain
        self.range = rng
        self.attrs = attrs
        self.expression = expression
        self.inv_expression = inv_expression
        self.set_points()

    def __str__(self):
        """A string version of the calibration."""
        result = ""
        result += "expression:\n"
        lines = str(self.expression).split("\n")
        for line in lines:
            result += " " * 4 + line + "\n"
        result += "params (p):\n"
        result += " " * 4 + str(self.params) + "\n"
        if len(self.aux_params) > 0:
            result += "auxiliary params (a):\n"
            result += " " * 4 + str(self.aux_params) + "\n"
        if self.inv_expression is not None:
            result += "inv_expression:\n"
            lines = str(self.inv_expression).split("\n")
            for line in lines:
                result += " " * 4 + line + "\n"
        result += "domain:\n"
        result += " " * 4 + str(self.domain) + "\n"
        result += "range:\n"
        result += " " * 4 + str(self.range) + "\n"
        if len(self.points_x) > 0:
            result += "calibration points (x):\n"
            result += " " * 4 + str(self.points_x) + "\n"
            result += "calibration points (y):\n"
            result += " " * 4 + str(self.points_y) + "\n"
            result += "calibration weights:\n"
            result += " " * 4 + str(self.weights) + "\n"
        if len(self.attrs.keys()) > 0:
            result += "other attributes:\n"
            result += " " * 4 + str(self.attrs)
        return result

    def __repr__(self):
        """A string representation of the calibration."""
        result = "Calibration("
        result += repr(self.expression)
        result += ", "
        result += repr(self.params)
        if len(self.aux_params) > 0:
            result += ", aux_params="
            result += repr(self.aux_params)
        if self.inv_expression is not None:
            result += ", inv_expression="
            result += repr(self.inv_expression)
        result += ", domain="
        result += repr(self.domain)
        result += ", rng="
        result += repr(self.range)
        if len(self.attrs) > 0:
            for key in self.attrs:
                result += ", "
                result += f"{key}={repr(self.attrs[key])}"
        result += ")"
        return result

    @property
    def expression(self):
        return self._expression

    @expression.setter
    def expression(self, expression):
        expression = _validate_expression(
            expression, domain=self.domain, rng=self.range
        )
        self._expression = expression

    @property
    def inv_expression(self):
        return self._inv_expression

    @inv_expression.setter
    def inv_expression(self, inv_expression):
        if inv_expression is None:
            self._inv_expression = inv_expression
        else:
            inv_expression = _validate_expression(
                inv_expression, ind_var="y", domain=self.range, rng=self.domain
            )
            self._inv_expression = inv_expression

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, p):
        params = np.array(p)
        if params.ndim != 1:
            raise CalibrationError(f"Parameters must be a 1-D array: {params}")
        self._params = params

    @property
    def aux_params(self):
        return self._aux_params

    @aux_params.setter
    def aux_params(self, a):
        if a is None:
            self._aux_params = np.array([])
        else:
            self._aux_params = np.array(a)

    @property
    def domain(self):
        return self._domain

    @domain.setter
    def domain(self, domain):
        _validate_domain_range(domain, (0, 1))
        self._domain = tuple(domain)

    @property
    def range(self):
        return self._range

    @range.setter
    def range(self, rng):
        _validate_domain_range((0, 1), rng)
        self._range = tuple(rng)

    @property
    def attrs(self):
        return self._attrs

    @attrs.setter
    def attrs(self, attrs):
        self._attrs = dict(copy.deepcopy(attrs))

    @property
    def points_x(self):
        return self._points_x

    @property
    def points_y(self):
        return self._points_y

    @property
    def weights(self):
        return self._weights

    @property
    def fit_y(self):
        """Calibration evaluated at the input x values."""
        return self(self.points_x)

    def add_points(self, points_x=None, points_y=None, weights=None):
        """Add the calibration point values to the internal list.

        Parameters
        ----------
        points_x : float or array_like
            The x-value or values of calibration points
        points_y : float or array_like
            The y-value or values of calibration points
        weights : float or array_like
            The weights to be used when fitting. Fitting will minimize the sum
            of squared errors weighted by these weights, i.e.,
            sum_i (w_i*(y_i - f(x_i))**2)). By default will use an array of
            ones with the same length as the points.
        """
        points_x, points_y, weights = _check_points(
            points_x, points_y, weights=weights, domain=self.domain, rng=self.range
        )
        self._points_x = np.append(self._points_x, points_x)
        self._points_y = np.append(self._points_y, points_y)
        self._weights = np.append(self._weights, weights)
        self._points_x, self._points_y, self._weights = _check_points(
            self._points_x,
            self._points_y,
            weights=self._weights,
            domain=self.domain,
            rng=self.range,
        )

    def set_points(self, points_x=None, points_y=None, weights=None):
        """Remove existing points and set the calibration point values.

        Parameters
        ----------
        points_x : float or array_like
            The x-value or values of calibration points
        points_y : float or array_like
            The y-value or values of calibration points
        weights : float or array_like
            The weights to be used when fitting. Fitting will minimize the sum
            of squared errors weighted by these weights, i.e.,
            sum_i (w_i*(y_i - f(x_i))**2)). By default will use an array of
            ones with the same length as the points.
        """
        self._points_x = []
        self._points_y = []
        self._weights = []
        self.add_points(points_x=points_x, points_y=points_y, weights=weights)

    def __eq__(self, other):
        """Determine if the two calibrations are identical."""
        if not isinstance(other, Calibration):
            raise CalibrationError(
                f"Attempting to compare {self.__class__} and {other.__class__}"
            )
        if len(self.params) != len(other.params):
            return False
        return (
            (self.expression == other.expression)
            and np.allclose(self.params, other.params)
            and np.allclose(self.aux_params, other.aux_params)
        )

    def copy(self):
        """Make a complete copy of the calibration."""
        cal = Calibration(
            self.expression,
            self.params,
            aux_params=self.aux_params,
            inv_expression=self.inv_expression,
            domain=self.domain,
            rng=self.range,
            **self.attrs,
        )
        cal.set_points(cal.points_x, cal.points_y)
        return cal

    def __call__(self, x):
        """Call the calibration function.

        Parameters
        ----------
        x : float or array_like
            The scalar argument(s) to the function (e.g., raw channel).

        Returns
        -------
        calibration : float or np.ndarray
            The value(s) of the calibration function at x.
        """
        return _eval_expression(
            self.expression,
            self.params,
            x,
            aux_params=self.aux_params,
            domain=self.domain,
            rng=self.range,
        )

    def inverse(self, y, x0=None, **kwargs):
        """Call the inverse of the calibration function.

        Parameters
        ----------
        y : float or array_like
            The value of the calibration function that we want to find the
            argument for.
        x0 : float or array_like
            A guess of the inverse of the function. If inverse expression
            does not exist, this can speed up the process of calculating
            the inverse.
        kwargs : dict
            Kwargs to be sent to scipy.optimize.root_scalar.

        Returns
        -------
        x : float or np.ndarray
            The argument of the inverse of the calibration function at y.
        """
        if np.any((y < self.range[0]) | (self.range[1] < y)):
            raise CalibrationError(f"Value {y} is outside the range {self.range}")
        if self.inv_expression is None:
            bracket = self.domain
            if isinstance(y, float):
                result = scipy.optimize.root_scalar(
                    lambda _x: self(_x) - y, x0=x0, bracket=bracket, **kwargs
                )
                x = result.root
            else:
                y = np.asarray(y)
                x = np.zeros_like(y)
                if x0 is not None:
                    x0 = np.asarray(x0)
                for j in range(len(x)):
                    if x0 is not None:
                        x0j = x0[j]
                    else:
                        x0j = None
                    result = scipy.optimize.root_scalar(
                        lambda _x: self(_x) - y[j], x0=x0j, bracket=bracket, **kwargs
                    )
                    x[j] = result.root
        else:
            x = _eval_expression(
                self.inv_expression,
                self.params,
                y,
                ind_var="y",
                aux_params=self.aux_params,
                domain=self.range,  # domain and range are swapped for inverse
                rng=self.domain,
            )
        # perform a final check on the calculated inverse
        if not np.allclose(self(x), y):
            raise CalibrationError(
                f"Inverse function did not result in matching y values:\n{self(x)}\n{y}"
            )
        return x

    @classmethod
    def read(cls, name):
        """Read the class from HDF5.

        Parameters
        ----------
        name : str, h5py.File, h5py.Group
            The filename or an open h5py File or Group.

        Returns
        -------
        calibration : becquerel.Calibration
        """
        dsets, attrs, skipped = io.h5.read_h5(name)
        if "params" not in dsets:
            raise CalibrationError('Expected dataset "params"')
        if "expression" not in dsets:
            raise CalibrationError('Expected dataset "expression"')
        if "domain" not in dsets:
            raise CalibrationError('Expected dataset "domain"')
        if "range" not in dsets:
            raise CalibrationError('Expected dataset "range"')
        unexpected = set(dsets.keys()) - {
            "expression",
            "params",
            "aux_params",
            "inv_expression",
            "domain",
            "range",
            "points_x",
            "points_y",
            "weights",
        }
        if len(unexpected) > 0:
            raise CalibrationError(f"Unexpected dataset names in file: {unexpected}")
        if "aux_params" in dsets:
            aux_params = dsets["aux_params"][()]
        else:
            aux_params = None
        expr = io.h5.ensure_string(dsets["expression"])
        if "inv_expression" in dsets:
            inv_expr = io.h5.ensure_string(dsets["inv_expression"])
        else:
            inv_expr = None
        cal = cls(
            expr,
            dsets["params"],
            aux_params=aux_params,
            inv_expression=inv_expr,
            domain=dsets["domain"],
            rng=dsets["range"],
            **attrs,
        )
        if "points_x" in dsets and "points_y" in dsets and "weights" in dsets:
            cal.set_points(dsets["points_x"], dsets["points_y"], dsets["weights"])
        elif "points_x" in dsets and "points_y" in dsets:
            cal.set_points(dsets["points_x"], dsets["points_y"])
        for key in attrs:
            if isinstance(attrs[key], (str, bytes)):
                attrs[key] = io.h5.ensure_string(attrs[key])
        return cal

    def write(self, name):
        """Write the class to HDF5.

        Parameters
        ----------
        name : str, h5py.File, h5py.Group
            The filename or an open h5py File or Group.
        """
        dsets = {
            "expression": self.expression,
            "params": self.params,
            "domain": self.domain,
            "range": self.range,
            "points_x": self.points_x,
            "points_y": self.points_y,
            "weights": self.weights,
        }
        if len(self.aux_params) > 0:
            dsets["aux_params"] = self.aux_params
        if self.inv_expression is not None:
            dsets["inv_expression"] = self.inv_expression
        attrs = copy.deepcopy(self.attrs)
        io.h5.write_h5(name, dsets, attrs)

    def fit(self, **kwargs):
        """Fit the calibration to the stored calibration points.

        Parameters
        ----------
        kwargs : dict
            Kwargs to pass to the minimization routine.
        """
        params = _fit_expression(
            self.expression,
            self.points_x,
            self.points_y,
            weights=self.weights,
            params0=self.params,
            aux_params=self.aux_params,
            domain=self.domain,
            rng=self.range,
            **kwargs,
        )
        self.params = params

    def fit_points(self, points_x, points_y, weights=None, params0=None, **kwargs):
        """Set the calibration point values and fit them.

        Convenience function for calling set_points() and fit().

        Parameters
        ----------
        points_x : float or array_like
            The x-value or values of calibration points
        points_y : float or array_like
            The y-value or values of calibration points
        weights : float or array_like
            The weights to be used when fitting. Fitting will minimize the sum
            of squared errors weighted by these weights, i.e.,
            sum_i (w_i*(y_i - f(x_i))**2)). By default will use an array of
            ones with the same length as the points.
        params0 : float or array_like
            Initial guesses for the parameters. By default an array of ones
            with its length inferred from the number of parameters
            referenced in the expression.
        kwargs : dict
            Kwargs to pass to the minimization routine.
        """
        self.set_points(points_x=points_x, points_y=points_y, weights=weights)
        if params0 is not None:
            self.params = params0
        self.fit(**kwargs)

    @classmethod
    def from_points(
        cls,
        expression,
        points_x,
        points_y,
        weights=None,
        params0=None,
        aux_params=None,
        domain=DEFAULT_DOMAIN,
        rng=DEFAULT_RANGE,
        fit_kwargs={},
        **attrs,
    ):
        """Create a Calibration with the expression and fit the points.

        Parameters
        ----------
        expression : string
            The expression that defines the calibration function.
        points_x : float or array_like
            The x-value or values of calibration points
        points_y : float or array_like
            The y-value or values of calibration points
        weights : float or array_like
            The weights to be used when fitting. Fitting will minimize the sum
            of squared errors weighted by these weights, i.e.,
            sum_i (w_i*(y_i - f(x_i))**2)). By default will use an array of
            ones with the same length as the points.
        params0 : float or array_like
            Initial guesses for the parameters. By default an array of ones
            with its length inferred from the number of parameters
            referenced in the expression.
        aux_params : array_like
            Auxiliary floating point parameters for the calibration function,
            referred to by the symbol "a". By default an empty array.
        domain : array_like
            The domain of the function. Will raise an error if the independent
            variable is outside this interval. Must be finite. By default
            DEFAULT_DOMAIN.
        rng : array_like
            The range of the function. Expression outputs will be clipped to
            this interval. Must be finite. By default DEFAULT_RANGE.
        fit_kwargs : dict
            Kwargs to pass to the minimization routine.
        attrs : dict
            Other information to be stored with the calibration.

        Returns
        -------
        cal : Calibration
            The Calibration instance with the given expression fitted to
            the points.
        """
        points_x, points_y, weights = _check_points(
            points_x, points_y, weights=weights, domain=domain, rng=rng
        )
        params = _fit_expression(
            expression,
            points_x,
            points_y,
            weights=weights,
            params0=params0,
            aux_params=aux_params,
            domain=domain,
            rng=rng,
            **fit_kwargs,
        )
        cal = cls(
            expression, params, aux_params=aux_params, domain=domain, rng=rng, **attrs
        )
        cal.set_points(points_x, points_y, weights)
        return cal

    @classmethod
    def from_linear(cls, params, domain=DEFAULT_DOMAIN, rng=DEFAULT_RANGE, **attrs):
        """Create a Calibration with a linear function.

        Parameters
        ----------
        params : array_like
            Coefficients beginning with 0th order.
        attrs : dict
            Other information to be stored with the calibration.
        """
        expr = "p[0] + p[1] * x"
        if len(params) != 2:
            raise CalibrationError("Linear calibration expects 2 parameters")
        return cls(expr, params, domain=domain, rng=rng, **attrs)

    @classmethod
    def from_polynomial(cls, params, domain=DEFAULT_DOMAIN, rng=DEFAULT_RANGE, **attrs):
        """Create a Calibration with a polynomial function of any order.

        The calibration function expression is
            "p[0] + p[1] * x + p[2] * x**2 + ..."

        Parameters
        ----------
        params : array_like
            Coefficients beginning with 0th order.
        attrs : dict
            Other information to be stored with the calibration.
        """
        expr = _polynomial_expression(params)
        return cls(expr, params, domain=domain, rng=rng, **attrs)

    @classmethod
    def from_sqrt_polynomial(
        cls, params, domain=DEFAULT_DOMAIN, rng=DEFAULT_RANGE, **attrs
    ):
        """Create a square root of a polynomial function of any order.

        The calibration function expression is
            "np.sqrt(p[0] + p[1] * x + p[2] * x**2 + ...)"

        Parameters
        ----------
        params : array_like
            Coefficients beginning with 0th order.
        attrs : dict
            Other information to be stored with the calibration.
        """
        expr = _polynomial_expression(params)
        expr = "np.sqrt(" + expr + ")"
        return cls(expr, params, domain=domain, rng=rng, **attrs)

    @classmethod
    def from_interpolation(
        cls, points_x, points_y, domain=DEFAULT_DOMAIN, rng=DEFAULT_RANGE, **attrs
    ):
        """Create a Calibration that interpolates the calibration points.

        Parameters
        ----------
        points_x : float or array_like
            The x-value or values of calibration points
        points_y : float or array_like
            The y-value or values of calibration points
        attrs : dict
            Other information to be stored with the calibration.
        """
        points_x, points_y, _ = _check_points(
            points_x, points_y, domain=domain, rng=rng
        )
        if len(points_x) < 2:
            raise CalibrationError("Interpolated calibration expects at least 2 points")
        # place the points into array of auxiliary params "a"
        aux_params = np.array([points_x, points_y])
        expr = ""
        expr += (
            f"scipy.interpolate.interp1d(a[0, :], a[1, :], fill_value='extrapolate')(x)"
        )
        return cls(expr, [], aux_params=aux_params, domain=domain, rng=rng, **attrs)

    @property
    def fit_R_squared(self):
        """Calibration fit R^2 value.

        Reference
        ---------
        stackoverflow.com/questions/19189362
        """

        # residual sum of squares
        ss_res = np.sum(self.weights * (self.points_y - self.fit_y) ** 2)

        # total sum of squares
        ss_tot = np.sum(self.weights * (self.points_y - np.mean(self.points_y)) ** 2)

        # r-squared
        return 1 - (ss_res / ss_tot)

    @property
    def fit_chi_squared(self):
        """Calibration fit chi^2 value."""

        if self.points_y.shape != self.fit_y.shape:
            raise ValueError(
                "y and fit_y must have same shapes:", self.y.shape, self.fit_y.shape
            )
        # Mask out zeros
        fit_y = self.fit_y[self.points_y > 0]
        points_y = self.points_y[self.points_y > 0]
        weights = self.weights[self.points_y > 0]
        return np.sum(weights * (points_y - fit_y) ** 2)

    @property
    def fit_degrees_of_freedom(self):
        """Calibration fit number of degrees of freedom."""
        return len(self.points_x) - len(self.params)

    @property
    def fit_reduced_chi_squared(self):
        """Calibration fit reduced chi^2 value."""
        return self.fit_chi_squared / self.fit_degrees_of_freedom

    def plot(self, ax=None):
        """Plot the calibration, and residuals if points exist.

        Parameters
        ----------
        ax : np.ndarray of shape (2,), or matplotlib axes object, optional
            Plot axes to use. If None, create new axes.
        """

        # Handle whether we have fit points or just a fit function
        has_points = self.points_x.size > 0

        if ax is None:
            fig, ax = plt.subplots(1 + has_points, 1, sharex=True)

        if has_points:
            assert ax.shape == (2,)
            ax_cal, ax_res = ax
            xmin, xmax = self.points_x.min(), self.points_x.max()
        else:
            ax_cal = ax
            xmin, xmax = self.domain

        # Plot calibration curve
        xx = np.linspace(xmin, xmax, 1000)
        yy = self(xx)
        ax_cal.plot(xx, yy, alpha=1.0 - 0.7 * has_points)
        ax_cal.set_ylabel("$y$")

        if has_points:
            # Plot calibration points
            ax_cal.scatter(self.points_x, self.points_y)

            # Plot residuals
            ax_res.scatter(self.points_x, self(self.points_x) - self.points_y)
            ax_res.set_xlabel("$x$")
            ax_res.set_ylabel("$y-x$")
            ax_res.axhline(0, linestyle="dashed", linewidth=1, c="k")
        else:
            ax_cal.set_xlabel("x")



"""Read in a becquerel Spectrum HDF5 file."""

from .. import io
from ..core import calibration
from .parsers import BecquerelParserError


def read(filename, verbose=False):
    """Parse the becquerel Spectrum HDF5 file and return a dictionary of data.

    Parameters
    ----------
    filename : str
        The filename of the HDF5 file to read.
    verbose : bool (optional)
        Whether to print out debugging information. By default False.

    Returns
    -------
    data : dict
        Dictionary of data that can be used to instantiate a Spectrum.
    cal : Calibration
        Energy calibration stored in the file.
    """
    print("Reading HDF5 file " + filename)
    if not io.h5.is_h5_filename(filename):
        raise BecquerelParserError("File is not an HDF5: " + filename)

    # group datasets and attributes into one dictionary
    dsets, attrs, skipped = io.h5.read_h5(filename)
    data = {**dsets, **attrs}

    # read energy calibration from file
    if "energy_cal" in skipped:
        with io.h5.open_h5(filename, "r") as h5:
            group = h5["energy_cal"]
            cal = calibration.Calibration.read(group)
    else:
        cal = None

    return data, cal



"""Becquerel: Tools for radiation spectral analysis."""

from .__metadata__ import (
    __description__,
    __url__,
    __version__,
    __license__,
    __copyright__,
)

from . import core
from .core import utils, fitting
from .core.autocal import AutoCalibrator, AutoCalibratorError
from .core.energycal import LinearEnergyCal, EnergyCalError, BadInput
from .core.calibration import Calibration, CalibrationError
from .core.fitting import Fitter
from .core.peakfinder import (
    PeakFilter,
    PeakFilterError,
    GaussianPeakFilter,
    PeakFinder,
    PeakFinderError,
)
from .core.plotting import SpectrumPlotter, PlottingError
from .core.rebin import rebin, RebinError, RebinWarning
from .core.spectrum import Spectrum, SpectrumError, UncalibratedError, SpectrumWarning
from .core.utils import UncertaintiesError

from . import parsers

from . import tools
from .tools import nndc, xcom, materials
from .tools.element import Element
from .tools.isotope import Isotope
from .tools.isotope_qty import IsotopeQuantity

import warnings

warnings.simplefilter("default", DeprecationWarning)

__all__ = [
    "__description__",
    "__url__",
    "__version__",
    "__license__",
    "__copyright__",
    "core",
    "utils",
    "fitting",
    "AutoCalibrator",
    "AutoCalibratorError",
    "LinearEnergyCal",
    "EnergyCalError",
    "BadInput",
    "Calibration",
    "CalibrationError",
    "Fitter",
    "PeakFilter",
    "PeakFilterError",
    "GaussianPeakFilter",
    "PeakFinder",
    "PeakFinderError",
    "SpectrumPlotter",
    "PlottingError",
    "rebin",
    "RebinError",
    "RebinWarning",
    "Spectrum",
    "SpectrumError",
    "UncalibratedError",
    "SpectrumWarning",
    "UncertaintiesError",
    "parsers",
    "tools",
    "nndc",
    "xcom",
    "materials",
    "Element",
    "Isotope",
    "IsotopeQuantity",
]



"""Query material data NIST X-Ray Mass Attenuation Coefficients website.

References:
  https://www.nist.gov/pml/x-ray-mass-attenuation-coefficients
  http://physics.nist.gov/PhysRefData/XrayMassCoef/tab1.html
  http://physics.nist.gov/PhysRefData/XrayMassCoef/tab2.html

"""

import requests
import pandas as pd
from collections.abc import Iterable
from .element import element_symbol

MAX_Z = 92
N_COMPOUNDS = 48


_URL_TABLE1 = "http://physics.nist.gov/PhysRefData/XrayMassCoef/tab1.html"
_URL_TABLE2 = "http://physics.nist.gov/PhysRefData/XrayMassCoef/tab2.html"


class NISTMaterialsError(Exception):
    """General error for NIST materials data."""

    pass


class NISTMaterialsRequestError(NISTMaterialsError):
    """Error related to communicating with NIST or parsing the result."""

    pass


def _get_request(url):
    """Perform GET request.

    Args:
      url: URL (string) to get.

    Returns:
      requests object.

    Raises:
      NISTMaterialsRequestError: if there was a problem making the request.

    """

    req = requests.get(url)
    if not req.ok or req.reason != "OK" or req.status_code != 200:
        raise NISTMaterialsRequestError(
            "NIST materials request failed: reason={}, status_code={}".format(
                req.reason, req.status_code
            )
        )
    return req


def fetch_element_data():
    """Retrieve data for the elements.

    Data are found in Table 1:
      http://physics.nist.gov/PhysRefData/XrayMassCoef/tab1.html

    Returns:
      A pandas DataFrame of the material data. Colums are 'Z', 'Symbol',
      'Element', 'Z_over_A', 'I_eV', and 'Density'.

    Raises:
      NISTMaterialsRequestError: if there was a problem obtaining the data.

    """
    req = _get_request(_URL_TABLE1)
    text = req.text
    # rename first two header columns
    text = text.replace(
        '<TH scope="col" COLSPAN="2"><I>Z</I></TH>',
        '<TH scope="col">Z</TH><TH scope="col">Symbol</TH>',
    )
    # remove row that makes the table too wide
    text = text.replace('<TD COLSPAN="10"><HR SIZE="1" NOSHADE></TD>', "")
    # remove extra header columns
    text = text.replace('<TD COLSPAN="2">', "<TD>")
    text = text.replace('<TD COLSPAN="4">', "<TD>")
    text = text.replace("TD>&nbsp;</TD>", "")
    # remove extra columns in Hydrogen row
    text = text.replace('<TD ROWSPAN="92">&nbsp;</TD>', "")
    # remove open <TR> at the end of the table
    text = text.replace("</TD></TR><TR>", "</TD></TR>")
    # read HTML table into pandas DataFrame
    tables = pd.read_html(text, header=0, skiprows=[1, 2])
    if len(tables) != 1:
        raise NISTMaterialsRequestError(
            f"1 HTML table expected, but found {len(tables)}"
        )
    df = tables[0]
    if len(df) != MAX_Z:
        raise NISTMaterialsRequestError(
            f"{MAX_Z} elements expected, but found {len(df)}"
        )
    if len(df.columns) != 6:
        raise NISTMaterialsRequestError(
            f"10 columns expected, but found {len(df.columns)} ({df.columns})"
        )
    # rename columns
    df.columns = ["Z", "Symbol", "Element", "Z_over_A", "I_eV", "Density"]
    return df


def convert_composition(comp):
    """Convert composition by Z into composition by symbol.

    Args:
      comp: a list of strings from the last column of Table 2, e.g.,
        ["1: 0.111898", "8: 0.888102"]

    Returns:
      A list of strings of composition by symbol, e.g.,
        ["H 0.111898", "O 0.888102"]

    Raises:
      NISTMaterialsRequestError: if there was a problem making the conversion.

    """
    comp_sym = []
    if not isinstance(comp, Iterable):
        raise NISTMaterialsRequestError(
            f"Compound must be an iterable of strings: {comp}"
        )
    for line in comp:
        if not isinstance(line, str):
            raise NISTMaterialsRequestError(
                f"Line must be a string type: {line} {type(line)}"
            )
        try:
            z, weight = line.split(":")
        except ValueError:
            raise NISTMaterialsRequestError(f"Unable to split compound line: {line}")
        try:
            z = int(z)
        except ValueError:
            raise NISTMaterialsRequestError(
                f"Unable to convert Z {z} to integer: {line}"
            )
        if z < 1 or z > MAX_Z:
            raise NISTMaterialsRequestError(f"Z {z} out of range [1, {line}]: {MAX_Z}")
        comp_sym.append(element_symbol(z) + " " + weight.strip())
    return comp_sym


def fetch_compound_data():
    """Retrieve data for the compounds.

    Data are found in Table 2:
      http://physics.nist.gov/PhysRefData/XrayMassCoef/tab2.html

    Returns:
      A pandas DataFrame of the material data. Columns are 'Material',
      'Z_over_A', 'I_eV', 'Density', 'Composition_Z', and 'Composition_symbol'.

    Raises:
      NISTMaterialsRequestError: if there was a problem obtaining the data.

    """
    req = _get_request(_URL_TABLE2)
    text = req.text
    # remove extra header columns
    text = text.replace('<TD ROWSPAN="2">&nbsp;</TD>', "")
    # remove extra rows in header row
    text = text.replace(' ROWSPAN="2"', "")
    # remove extra columns in third header row
    text = text.replace('<TD COLSPAN="9"><HR SIZE="1" NOSHADE></TD>', "")
    # remove extra columns in the first material row
    text = text.replace('<TD ROWSPAN="50"> &nbsp; </TD>', "")
    # replace <BR> symbols in composition lists with semicolons
    text = text.replace("<BR>", ";")
    # read HTML table into pandas DataFrame
    tables = pd.read_html(text, header=0, skiprows=[1, 2])
    if len(tables) != 1:
        raise NISTMaterialsRequestError(
            f"1 HTML table expected, but found {len(tables)}"
        )
    df = tables[0]
    if len(df) != N_COMPOUNDS:
        raise NISTMaterialsRequestError(
            f"{N_COMPOUNDS} compounds expected, but found {len(df)}"
        )
    if len(df.columns) != 5:
        raise NISTMaterialsRequestError(
            f"5 columns expected, but found {len(df.columns)} ({df.columns})"
        )
    # rename columns
    df.columns = ["Material", "Z_over_A", "I_eV", "Density", "Composition_Z"]
    # clean up Z composition
    df["Composition_Z"] = [
        [line.strip() for line in comp.split(";")] for comp in df["Composition_Z"]
    ]
    # create a column of compositions by symbol (for use with fetch_xcom_data)
    df["Composition_symbol"] = [
        convert_composition(comp) for comp in df["Composition_Z"]
    ]
    return df
